from gaphas.guide import GuidedItemHandleMoveMixin
from gaphas.handlemove import HandleMove, ItemHandleMove
from gaphas.move import Move
from gaphas.types import Pos

from gaphor.diagram.presentation import connect
from gaphor.diagram.tools.grayout import GrayOutLineHandleMoveMixin
from gaphor.UML.actions.activity import ActivityParameterNodeItem


@HandleMove.register(ActivityParameterNodeItem)
class ActivityParameterNodeItemHandleMove(
    GrayOutLineHandleMoveMixin, GuidedItemHandleMoveMixin, ItemHandleMove
):
    """We use a custom tool for moving parameter nodes.

    Parameter nodes always connect to their parent (Activity).
    """

    def connect(self, pos: Pos) -> None:
        item = self.item
        connect(item, self.handle, item.parent)
        self.view.model.request_update(item)


@Move.register(ActivityParameterNodeItem)
def activity_parameter_node_move(item, view):
    """We use a custom tool for moving parameter nodes.

    They can be moved while handle is connected and will automatically
    connect to a side of its parent.
    """
    return ActivityParameterNodeItemHandleMove(item, item.handles()[0], view)



"""The properties module allows Gaphor properties to be saved to the local file
system.

These are things like preferences.
"""


import ast
import hashlib
import logging
import os
import pprint
from pathlib import Path
from typing import Dict

from gi.repository import GLib

from gaphor.abc import Service
from gaphor.core import event_handler
from gaphor.core.modeling.event import ModelFlushed
from gaphor.event import ModelLoaded, ModelSaved, SessionCreated

log = logging.getLogger(__name__)


def get_config_dir() -> str:
    """Return the directory where the user's config is stored.

    This varies depending on platform.
    """

    config_dir = os.path.join(GLib.get_user_config_dir(), "gaphor")
    os.makedirs(config_dir, exist_ok=True)

    return config_dir


def get_cache_dir() -> str:
    """Return the directory where the user's cache is stored.

    This varies depending on platform.
    """

    cache_dir = os.path.join(GLib.get_user_cache_dir(), "gaphor")
    os.makedirs(cache_dir, exist_ok=True)

    return cache_dir


def file_hash(filename) -> str:
    return hashlib.blake2b(str(filename).encode("utf-8"), digest_size=24).hexdigest()


class PropertyChanged:
    """This event is triggered any time a property is changed.

    This event holds the property key, the current value, and the new
    value.
    """

    def __init__(self, key, old_value, new_value):
        self.key = key
        self.old_value = old_value
        self.new_value = new_value


_no_default = object()


class Properties(Service):
    """The Properties class holds a collection of application wide properties.

    Properties are persisted to the local file system.
    """

    def __init__(self, event_manager):
        self.event_manager = event_manager
        self.filename = os.path.join(get_cache_dir(), file_hash(""))
        self._properties: Dict[str, object] = {}

        event_manager.subscribe(self.on_model_loaded)
        event_manager.subscribe(self.on_model_saved)
        event_manager.subscribe(self.on_model_flushed)

        self.load()

    def shutdown(self):
        """Shutdown the properties service.

        This will ensure that all properties are saved.
        """
        self.save()

        self.event_manager.unsubscribe(self.on_model_loaded)
        self.event_manager.unsubscribe(self.on_model_saved)
        self.event_manager.unsubscribe(self.on_model_flushed)

    @event_handler(ModelLoaded, SessionCreated)
    def on_model_loaded(self, event):
        self.filename = os.path.join(get_cache_dir(), file_hash(event.filename or ""))
        self.load()

    @event_handler(ModelSaved)
    def on_model_saved(self, event):
        self.filename = os.path.join(get_cache_dir(), file_hash(event.filename))
        self.save()

    @event_handler(ModelFlushed)
    def on_model_flushed(self, event):
        if self.filename:
            self.save()

    def __call__(self, key, default=_no_default):
        """Retrieve the specified property.

        If the property doesn't exist, the default parameter is
        returned.  This defaults to _no_default.
        """

        return self.get(key, default)

    def load(self):
        """Load properties from a file.

        Resources are loaded like you do with a dict().
        """

        filename = self.filename

        if os.path.exists(filename) and os.path.isfile(filename):

            data = Path(filename).read_text()
            try:
                self._properties = ast.literal_eval(data)
            except SyntaxError:
                log.error("Invalid syntax in property file %s", filename)

            for key, value in list(self._properties.items()):
                self.event_manager.handle(PropertyChanged(key, None, value))

    def save(self):
        """Save all properties by calling save() on the properties storage
        backend."""

        if not self.filename:
            return

        with open(self.filename, "w", encoding="utf-8") as ofile:
            pprint.pprint(self._properties, ofile)

    def get(self, key: str, default=_no_default):
        """Locate a property.

        Resource should be the class of the resource to look for or a
        string. In case of a string the resource will be looked up in
        the GConf configuration.
        """

        try:
            return self._properties[key]
        except KeyError as e:
            if default is _no_default:
                raise KeyError(f'No resource with name "{key}"') from e

            self.set(key, default)
            return default

    def set(self, key, value):
        """Set a property to a specific value.

        No smart things are done with classes and class names (like the
        resource() method does).
        """

        properties = self._properties
        old_value = properties.get(key)

        if value != old_value:
            properties[key] = value
            self.event_manager.handle(PropertyChanged(key, old_value, value))



"""An ExecutionSpecification is defined by a white recrange overlaying the
lifeline.

  ,----------.
  | lifeline |
  `----------'
       | --- Lifeline
      ,+. -- ExecutionOccurrenceSpecification
      | | -- ExecutionSpecification
      `+' -- ExecutionOccurrentSpecification
       |

    ExecutionOccurrenceSpecification.covered <--> Lifeline.coveredBy
    ExecutionOccurrenceSpecification.execution <--> ExecutionSpecification.execution

NB. ExecutionSpecification is abstract. Should use either
ActionExecutionSpecification or BehaviorExecutionSpecification.
What's the difference?

Stick with BehaviorExecutionSpecification, since it has a [0..1] relation to
behavior, whereas ActionExecutionSpecification has a [1] relation to action.
"""
from gaphas import Handle
from gaphas.connector import Position
from gaphas.constraint import constraint
from gaphas.geometry import Rectangle, distance_rectangle_point
from gaphas.solver import STRONG

from gaphor import UML
from gaphor.core.modeling import Presentation
from gaphor.diagram.presentation import (
    HandlePositionUpdate,
    literal_eval,
    postload_connect,
)
from gaphor.diagram.shapes import Box, draw_border
from gaphor.diagram.support import represents
from gaphor.UML.interactions.lifeline import BetweenPort


@represents(UML.ExecutionSpecification)
@represents(UML.BehaviorExecutionSpecification)
class ExecutionSpecificationItem(
    Presentation[UML.ExecutionSpecification], HandlePositionUpdate
):
    """Representation of interaction execution specification."""

    def __init__(self, diagram, id=None):
        super().__init__(diagram, id=id)
        self._connections = diagram.connections

        self._bar_width = 12

        ht, hb = Handle(strength=STRONG), Handle(strength=STRONG)
        ht.connectable = True

        self._handles = [ht, hb]
        self.watch_handle(ht)
        self.watch_handle(hb)

        self._connections.add_constraint(self, constraint(vertical=(ht.pos, hb.pos)))

        r = self._bar_width / 2
        nw = Position(-r, 0, strength=STRONG)
        ne = Position(r, 0, strength=STRONG)
        se = Position(r, 0, strength=STRONG)
        sw = Position(-r, 0, strength=STRONG)

        for c in (
            constraint(horizontal=(nw, ht.pos)),
            constraint(horizontal=(ne, ht.pos)),
            constraint(horizontal=(sw, hb.pos)),
            constraint(horizontal=(se, hb.pos)),
            constraint(vertical=(nw, ht.pos), delta=r),
            constraint(vertical=(ne, ht.pos), delta=-r),
            constraint(vertical=(sw, hb.pos), delta=r),
            constraint(vertical=(se, hb.pos), delta=-r),
        ):
            self._connections.add_constraint(self, c)

        self._ports = [BetweenPort(nw, sw), BetweenPort(ne, se)]

        self._shape = Box(
            style={"background-color": (1.0, 1.0, 1.0, 1.0)}, draw=draw_border
        )

    def handles(self):
        return self._handles

    def ports(self):
        return self._ports

    @property
    def top(self):
        return self._handles[0]

    @property
    def bottom(self):
        return self._handles[1]

    def dimensions(self):
        d = self._bar_width
        pt, pb = (h.pos for h in self._handles)
        return Rectangle(pt.x - d / 2, pt.y, d, y1=pb.y)

    def draw(self, context):
        self._shape.draw(context, self.dimensions())

    def point(self, x, y):
        return distance_rectangle_point(self.dimensions(), (x, y))

    def save(self, save_func):
        def save_connection(name, handle):
            c = self._connections.get_connection(handle)
            if c:
                save_func(name, c.connected)

        points = [tuple(map(float, h.pos)) for h in self._handles]

        save_func("matrix", tuple(self.matrix))
        save_func("points", points)
        save_connection("head-connection", self._handles[0])
        super().save(save_func)

    def load(self, name, value):
        if name == "points":
            points = literal_eval(value)
            for h, p in zip(self._handles, points):
                h.pos = p
        elif name == "head-connection":
            self._load_head_connection = value
        else:
            super().load(name, value)
        if name == "points":
            # Fix ports, so connections are made properly
            pl, pr = self._ports
            ht, hb = self._handles
            r = self._bar_width / 2
            pl.start.y = pr.start.y = ht.pos.y
            pl.end.y = pr.end.y = hb.pos.y
            pl.start.x = pl.end.x = ht.pos.x - r
            pr.start.x = pr.end.x = ht.pos.x + r

    def postload(self):
        if hasattr(self, "_load_head_connection"):
            postload_connect(self, self._handles[0], self._load_head_connection)
            del self._load_head_connection

        super().postload()



import importlib.resources
from pathlib import Path
from typing import NamedTuple

from gi.repository import GLib, Gtk

from gaphor.abc import ActionProvider, Service
from gaphor.action import action
from gaphor.babel import translate_model
from gaphor.core import event_handler
from gaphor.diagram.hoversupport import flowbox_add_hover_support
from gaphor.event import ActiveSessionChanged, SessionCreated
from gaphor.i18n import gettext, translated_ui_string
from gaphor.ui import APPLICATION_ID


class ModelTemplate(NamedTuple):
    name: str
    icon: str
    lang: str
    filename: str


TEMPLATES = [
    ModelTemplate(gettext("Generic"), "org.gaphor.Gaphor", "UML", "blank.gaphor"),
    ModelTemplate(gettext("UML"), "UML", "UML", "uml.gaphor"),
    ModelTemplate(gettext("SysML"), "SysML", "SysML", "sysml.gaphor"),
    ModelTemplate(gettext("RAAML"), "RAAML", "RAAML", "raaml.gaphor"),
    ModelTemplate(gettext("C4 Model"), "C4Model", "C4Model", "c4model.gaphor"),
]


def new_builder(ui_file):
    builder = Gtk.Builder()
    ui_file = f"{ui_file}.glade" if Gtk.get_major_version() == 3 else f"{ui_file}.ui"
    builder.add_from_string(translated_ui_string("gaphor.ui", ui_file))
    return builder


class Greeter(Service, ActionProvider):
    def __init__(self, application, event_manager, recent_manager=None):
        self.templates = None
        self.application = application
        self.event_manager = event_manager
        self.recent_manager = recent_manager or Gtk.RecentManager.get_default()
        self.greeter: Gtk.Window = None
        self.stack: Gtk.Stack = None
        self.action_bar: Gtk.ActionBar = None
        self.back_button: Gtk.Button = None
        self.gtk_app: Gtk.Application = None
        event_manager.subscribe(self.on_session_created)
        self._on_recent_files_changed_id = 0

    def init(self, gtk_app):
        self.gtk_app = gtk_app
        self._on_recent_files_changed_id = self.recent_manager.connect(
            "changed", self._on_recent_files_changed
        )
        self._on_recent_files_changed()

    def shutdown(self):
        self.event_manager.unsubscribe(self.on_session_created)
        if self._on_recent_files_changed_id:
            self.recent_manager.disconnect(self._on_recent_files_changed_id)
            self._on_recent_files_changed_id = 0

        if self.greeter:
            self.greeter.destroy()
        self.gtk_app = None

    def open(self, stack_name=None) -> None:
        if self.greeter and self.stack:
            self.stack.set_visible_child_name(stack_name)
            self.greeter.present()
            return

        if not stack_name:
            stack_name = (
                "recent-files" if any(self.query_recent_files()) else "new-model"
            )

        builder = new_builder("greeter")

        listbox = builder.get_object("greeter-recent-files")
        listbox.connect("row-activated", self._on_recent_file_activated)
        for widget in self.create_recent_files():
            listbox.insert(widget, -1)

        self.action_bar = builder.get_object("action-bar")
        self.back_button = builder.get_object("back-button")

        templates = builder.get_object("templates")
        templates.connect("child-activated", self._on_template_activated)
        for widget in self.create_templates():
            templates.insert(widget, -1)

        if Gtk.get_major_version() == 3:
            flowbox_add_hover_support(templates)
        self.templates = templates

        self.stack = builder.get_object("stack")
        self.stack.set_visible_child_name(stack_name)
        self.stack.connect("notify::visible-child", self._on_stack_changed)

        self.greeter = builder.get_object("greeter")
        self.greeter.set_application(self.gtk_app)
        if Gtk.get_major_version() == 3:
            self.greeter.connect("delete-event", self._on_window_close_request)
        else:
            self.greeter.connect("close-request", self._on_window_close_request)

        self.set_widgets_visible()
        self.greeter.show()
        templates.unselect_all()

    def close(self):
        if self.greeter:
            self.greeter.destroy()
            self.greeter = None
            self.stack = None

    @action(name="app.recent-files", shortcut="<Primary>o")
    def recent_files(self):
        self.open("recent-files")

    @action(name="app.new-model", shortcut="<Primary>n")
    def new_model(self):
        self.open("new-model")

    def query_recent_files(self):
        for item in self.recent_manager.get_items():
            if APPLICATION_ID in item.get_applications() and item.exists():
                yield item

    def create_recent_files(self):
        for item in self.query_recent_files():
            builder = new_builder("greeter-recent-file")
            filename, _host = GLib.filename_from_uri(item.get_uri())
            builder.get_object("name").set_text(str(Path(filename).stem))
            builder.get_object("filename").set_text(
                item.get_uri_display().replace(str(Path.home()), "~")
            )
            row = builder.get_object("greeter-recent-file")
            row.filename = filename
            yield row

    def create_templates(self):
        for template in TEMPLATES:
            builder = new_builder("greeter-model-template")
            builder.get_object("template-name").set_text(template.name)
            if Gtk.get_major_version() == 3:
                builder.get_object("template-icon").set_from_icon_name(
                    template.icon, Gtk.IconSize.DIALOG
                )
            else:
                builder.get_object("template-icon").set_from_icon_name(template.icon)
            child = builder.get_object("model-template")
            child.filename = template.filename
            child.lang = template.lang
            yield child

    @event_handler(SessionCreated, ActiveSessionChanged)
    def on_session_created(self, _event=None):
        self.close()

    def _on_stack_changed(self, stack: Gtk.Stack, gparam_object):
        self.set_widgets_visible()

    def set_widgets_visible(self):
        visible = self.stack.get_visible_child_name()
        if visible == "new-model":
            self.action_bar.set_visible(False)
            if any(self.query_recent_files()):
                self.back_button.set_visible(True)
            else:
                self.back_button.set_visible(False)
            self.greeter.set_title(gettext("Create a New Model"))
        else:
            self.action_bar.set_visible(True)
            self.back_button.set_visible(False)
            self.greeter.set_title(gettext("Open a Recent Model"))

    def _on_recent_files_changed(self, recent_manager=None):
        self.gtk_app.lookup_action("recent-files").set_enabled(
            any(self.query_recent_files())
        )

    def _on_recent_file_activated(self, _listbox, row):
        filename = row.filename
        self.application.new_session(filename=filename)
        self.close()

    def _on_template_activated(self, _flowbox, child):
        filename: Path = (
            importlib.resources.files("gaphor") / "templates" / child.filename
        )
        translated_model = translate_model(filename)
        session = self.application.new_session(template=translated_model)
        session.get_service("properties").set("modeling-language", child.lang)
        self.close()

    def _on_window_close_request(self, window, event=None):
        self.close()
        return False



"""This is the TreeView that is most common (for example: it is used in
Rational Rose).

This is a tree based on namespace relationships. As a result only
classifiers are shown here.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from gaphas.decorators import g_async
from gi.repository import Gdk, Gio, GLib, Gtk

from gaphor import UML
from gaphor.abc import ActionProvider, ModelingLanguage
from gaphor.core import action, event_handler, gettext
from gaphor.core.modeling import Diagram, Element, Presentation
from gaphor.diagram.deletable import deletable
from gaphor.transaction import Transaction
from gaphor.ui.abc import UIComponent
from gaphor.ui.actiongroup import create_action_group
from gaphor.ui.event import DiagramOpened, DiagramSelectionChanged, ElementOpened
from gaphor.ui.mainwindow import create_diagram_types_model
from gaphor.ui.namespacemodel import (
    RELATIONSHIPS,
    NamespaceModel,
    NamespaceModelElementDropped,
    NamespaceModelRefreshed,
)
from gaphor.ui.namespaceview import namespace_view

if TYPE_CHECKING:
    from gaphor.core.eventmanager import EventManager
    from gaphor.core.modeling import ElementFactory


log = logging.getLogger(__name__)


def popup_model(element, modeling_language):
    model = Gio.Menu.new()

    part = Gio.Menu.new()

    part.append(
        gettext("_Open") if isinstance(element, Diagram) else gettext("Add to diagram"),
        "tree-view.open",
    )
    part.append(gettext("_Rename"), "tree-view.rename")
    model.append_section(None, part)

    part = Gio.Menu.new()
    part.append_submenu(
        gettext("New _Diagram"), create_diagram_types_model(modeling_language)
    )
    part.append(gettext("New _Package"), "tree-view.create-package")
    model.append_section(None, part)

    part = Gio.Menu.new()
    part.append(gettext("De_lete"), "tree-view.delete")
    model.append_section(None, part)

    if not element:
        return model

    part = Gio.Menu.new()
    for presentation in element.presentation:
        if diagram := presentation.diagram:
            menu_item = Gio.MenuItem.new(
                gettext("Show in “{diagram}”").format(diagram=diagram.name),
                "tree-view.show-in-diagram",
            )
            menu_item.set_attribute_value("target", GLib.Variant.new_string(diagram.id))
            part.append_item(menu_item)

        # Play it safe with an (arbitrary) upper bound
        if part.get_n_items() > 29:
            break

    if part.get_n_items() > 0:
        model.append_section(None, part)
    return model


class Namespace(UIComponent, ActionProvider):
    def __init__(
        self,
        event_manager: EventManager,
        element_factory: ElementFactory,
        modeling_language: ModelingLanguage,
    ):
        self.event_manager = event_manager
        self.element_factory = element_factory
        self.modeling_language = modeling_language

        self.model: NamespaceModel | None = None
        self.view: Gtk.TreeView | None = None
        self.scrolled_window: Gtk.ScrolledWindow | None = None
        self.ctrl: set[Gtk.EventController] = set()

    def open(self):
        self.model = NamespaceModel(self.event_manager, self.element_factory)
        view = namespace_view(self.model)
        self.event_manager.subscribe(self._on_model_refreshed)
        self.event_manager.subscribe(self._on_model_dropped)
        self.event_manager.subscribe(self._on_diagram_selection_changed)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        if Gtk.get_major_version() == 3:
            scrolled_window.add(view)
        else:
            scrolled_window.set_child(view)
        scrolled_window.show()
        view.show()

        view.connect("row-activated", self._on_view_row_activated)
        view.connect_after("cursor-changed", self._on_view_cursor_changed)
        view.connect("destroy", self._on_view_destroyed)

        action_group, shortcuts = create_action_group(self, "tree-view")
        scrolled_window.insert_action_group("tree-view", action_group)

        if Gtk.get_major_version() == 3:
            self.create_gtk3_popup_controller(view)
        else:
            self.create_gtk4_popup_controller(shortcuts, view)
        self.view = view
        self.scrolled_window = scrolled_window
        self.model.refresh()

        return scrolled_window

    def create_gtk4_popup_controller(self, shortcuts, view):
        ctrl = Gtk.ShortcutController.new_for_model(shortcuts)
        ctrl.set_scope(Gtk.ShortcutScope.LOCAL)
        view.add_controller(ctrl)

        ctrl = Gtk.GestureClick.new()
        ctrl.set_button(Gdk.BUTTON_SECONDARY)
        ctrl.connect("pressed", self._on_show_popup)
        view.add_controller(ctrl)

    def create_gtk3_popup_controller(self, view):
        ctrl = Gtk.GestureMultiPress.new(view)
        ctrl.set_button(Gdk.BUTTON_SECONDARY)
        ctrl.connect("pressed", self._on_show_popup)
        self.ctrl.add(ctrl)

        ctrl = Gtk.EventControllerKey.new(view)
        ctrl.connect("key-pressed", self._on_edit_pressed)
        self.ctrl.add(ctrl)

    def close(self):
        self.ctrl.clear()
        if self.view:
            if Gtk.get_major_version() == 3:
                self.view.destroy()
            elif self.scrolled_window:
                self.scrolled_window.unparent()
            self.view = None
            self.scrolled_window = None
        if self.model:
            self.model.shutdown()
            self.model = None
        self.event_manager.unsubscribe(self._on_model_refreshed)
        self.event_manager.unsubscribe(self._on_model_dropped)
        self.event_manager.unsubscribe(self._on_diagram_selection_changed)

    @event_handler(NamespaceModelRefreshed)
    def _on_model_refreshed(self, event):
        # Expand all root elements:
        if self.view:
            self.view.expand_row(path=Gtk.TreePath.new_first(), open_all=False)
            self._on_view_cursor_changed(self.view)

    @event_handler(NamespaceModelElementDropped)
    def _on_model_dropped(self, event):
        element = event.element
        self.select_element(element)

    @event_handler(DiagramSelectionChanged)
    def _on_diagram_selection_changed(self, event):
        focused_item = event.focused_item
        if isinstance(focused_item, Presentation) and focused_item.subject:
            self.select_element(focused_item.subject)

    def _on_show_popup(self, ctrl, n_press, x, y):
        selected_element = self.get_selected_element()
        if selected_element in (None, RELATIONSHIPS):
            return

        if Gtk.get_major_version() == 3:
            menu = Gtk.Menu.new_from_model(
                popup_model(selected_element, self.modeling_language)
            )
            menu.attach_to_widget(self.view, None)
            menu.popup_at_pointer(None)
        else:
            menu = Gtk.PopoverMenu.new_from_model(
                popup_model(selected_element, self.modeling_language)
            )
            gdk_rect = Gdk.Rectangle()
            gdk_rect.x = x
            gdk_rect.y = y
            gdk_rect.width = gdk_rect.height = 1

            menu.set_pointing_to(gdk_rect)
            menu.set_has_arrow(False)
            menu.set_parent(self.view)
            menu.popup()

    def _on_edit_pressed(self, ctrl, keyval, keycode, state):
        if keyval == Gdk.KEY_F2:
            self.tree_view_rename_selected()
            return True
        return False

    def _on_view_row_activated(self, view, path, column):
        """Double-click on an element in the tree view."""
        if Gtk.get_major_version() == 3:
            view.get_action_group("tree-view").lookup_action("open").activate()
        else:
            view.activate_action("tree-view.open", None)

    def _on_view_cursor_changed(self, view):
        """Another row is selected, toggle action sensitivity."""
        element = self.get_selected_element()
        if Gtk.get_major_version() == 3:
            action_group = view.get_action_group("tree-view")

            action_group.lookup_action("create-package").set_enabled(
                isinstance(element, UML.Package)
            )
            action_group.lookup_action("delete").set_enabled(
                element and deletable(element)
            )
            action_group.lookup_action("rename").set_enabled(
                isinstance(element, (Diagram, UML.NamedElement))
            )

    def _on_view_destroyed(self, widget):
        self.close()

    def select_element(self, element: Element) -> None:
        """Select an element from the Namespace view.

        The element is selected. After this an action may be executed,
        such as OpenModelElement, which will try to open the element (if
        it's a Diagram).
        """
        assert self.model
        assert self.view

        model = self.view.get_model()
        tree_iter = self.model.iter_for_element(element)
        if not tree_iter:
            return

        path = model.get_path(tree_iter)
        path_indices = path.get_indices()

        # Expand the parent row
        if len(path_indices) > 1:
            parent_path = Gtk.TreePath.new_from_indices(path_indices[:-1])
            self.view.expand_row(path=parent_path, open_all=False)

        self.view.set_cursor(path, self.view.get_column(0), False)
        self.view.scroll_to_cell(path, None, False, 0, 0)
        self._on_view_cursor_changed(self.view)

    def get_selected_element(self) -> Element | None:
        assert self.view
        assert self.model
        path, _column = self.view.get_cursor()
        if not path:
            return None
        model = self.model
        return model.get_value(model.get_iter(path), 0)  # type: ignore[no-any-return]

    @action(name="tree-view.open")
    def tree_view_open_selected(self):
        element = self.get_selected_element()
        if isinstance(element, Diagram):
            self.event_manager.handle(DiagramOpened(element))
        else:
            self.event_manager.handle(ElementOpened(element))

    @action(name="tree-view.show-in-diagram")
    def tree_view_show_in_diagram(self, diagram_id: str) -> None:
        element = self.element_factory.lookup(diagram_id)
        self.event_manager.handle(DiagramOpened(element))

    @action(name="tree-view.rename", shortcut="F2")
    @g_async(single=True)
    def tree_view_rename_selected(self):
        assert self.view
        view = self.view
        element = self.get_selected_element()
        path, column = view.get_cursor()
        if path and column and element not in (None, RELATIONSHIPS):
            cell = column.get_cells()[1]
            cell.set_property("editable", 1)
            view.set_cursor(path, column, True)
            cell.set_property("editable", 0)

    @action(name="win.create-diagram")
    def tree_view_create_diagram(self, diagram_type: str):
        assert self.view
        element = self.get_selected_element()
        while element and not isinstance(element, UML.NamedElement):
            element = element.owner

        with Transaction(self.event_manager):
            diagram = self.element_factory.create(Diagram)
            if isinstance(element, UML.NamedElement):
                diagram.element = element
            diagram.name = diagram_name_for_type(self.modeling_language, diagram_type)
            diagram.diagramType = diagram_type
        self.select_element(diagram)
        self.event_manager.handle(DiagramOpened(diagram))
        self.tree_view_rename_selected()

    @action(name="tree-view.create-package")
    def tree_view_create_package(self):
        element = self.get_selected_element()
        assert isinstance(element, UML.Package)
        with Transaction(self.event_manager):
            package = self.element_factory.create(UML.Package)
            package.package = element
            package.name = (
                gettext("{name} package").format(name=element.name)
                if element
                else gettext("New model")
            )
        self.select_element(package)
        self.tree_view_rename_selected()

    @action(name="tree-view.delete")
    def tree_view_delete(self):
        if element := self.get_selected_element():
            with Transaction(self.event_manager):
                element.unlink()


def diagram_name_for_type(modeling_language, diagram_type):
    for id, name, _ in modeling_language.diagram_types:
        if id == diagram_type:
            return name
    return gettext("New diagram")



"""Support code for dealing with stereotypes in diagrams."""

from gaphor.core.format import format
from gaphor.core.styling import JustifyContent, TextAlign
from gaphor.diagram.shapes import Box, Text, draw_top_separator


def stereotype_watches(presentation):
    presentation.watch("subject.appliedStereotype", presentation.update_shapes).watch(
        "subject.appliedStereotype.classifier.name"
    ).watch("subject.appliedStereotype.slot", presentation.update_shapes).watch(
        "subject.appliedStereotype.slot.definingFeature.name"
    ).watch(
        "subject.appliedStereotype.slot.value", presentation.update_shapes
    )


def stereotype_compartments(subject):
    return filter(
        None,
        (
            _create_stereotype_compartment(appliedStereotype)
            for appliedStereotype in subject.appliedStereotype
        )
        if subject
        else [],
    )


def _create_stereotype_compartment(appliedStereotype):
    def lazy_format(slot):
        return lambda: format(slot)

    slots = [slot for slot in appliedStereotype.slot if slot.value]

    if slots:
        return Box(
            Text(
                text=lazy_format(appliedStereotype.classifier[0]),
                style={"padding": (0, 0, 4, 0)},
            ),
            *(
                Text(text=lazy_format(slot), style={"text-align": TextAlign.LEFT})
                for slot in slots
            ),
            style={
                "padding": (4, 4, 4, 4),
                "min-height": 8,
                "justify-content": JustifyContent.START,
            },
            draw=draw_top_separator,
        )
    else:
        return None



from gaphor.C4Model import c4model
from gaphor.core import gettext
from gaphor.core.styling import FontWeight, JustifyContent, TextAlign
from gaphor.diagram.presentation import ElementPresentation, Named
from gaphor.diagram.shapes import Box, Text, draw_border
from gaphor.diagram.support import represents


@represents(c4model.C4Container)
class C4ContainerItem(Named, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)

        self.watch("subject[NamedElement].name")
        self.watch("subject[C4Container].technology")
        self.watch("subject[C4Container].description")
        self.watch("subject[C4Container].type")
        self.watch("children", self.update_shapes)

    def update_shapes(self, event=None):
        text_align = (
            TextAlign.LEFT if self.diagram and self.children else TextAlign.CENTER
        )
        self.shape = Box(
            Text(
                text=lambda: self.subject.name or "",
                style={"font-weight": FontWeight.BOLD, "text-align": text_align},
            ),
            Text(
                text=lambda: self.subject.technology
                and f"[{gettext(self.subject.type)}: {self.subject.technology}]"
                or f"[{gettext(self.subject.type)}]",
                style={"font-size": "x-small", "text-align": text_align},
            ),
            *(
                ()
                if self.children
                else (
                    Text(
                        text=lambda: self.subject.description or "",
                        width=lambda: self.width - 8,
                        style={"padding": (4, 0, 0, 0), "text-align": text_align},
                    ),
                )
            ),
            style={
                "padding": (4, 4, 4, 4),
                "justify-content": JustifyContent.END
                if self.diagram and self.children
                else JustifyContent.CENTER,
            },
            draw=draw_border,
        )



from gaphor import UML
from gaphor.diagram.group import group, ungroup


@group.register(UML.Package, UML.Type)
@group.register(UML.Package, UML.Package)
def container_group(parent, element) -> bool:
    if element.owner:
        ungroup(element.owner, element)

    element.package = parent
    return True


@ungroup.register(UML.Package, UML.Type)
@ungroup.register(UML.Package, UML.Package)
def container_ungroup(parent, element) -> bool:
    if element.package is parent:
        del element.package
        return True
    return False


@group.register(UML.Class, UML.Class)
def class_group(parent, element) -> bool:
    if element.owner:
        ungroup(element.owner, element)
    element.nestingClass = parent
    return True


@ungroup.register(UML.Class, UML.Class)
def class_ungroup(parent, element) -> bool:
    if element.nestingClass is parent:
        del element.nestingClass
        return True
    return False


@group.register(UML.BehavioredClassifier, UML.Behavior)
def behavior_group(parent, element) -> bool:
    if element.owner:
        ungroup(element.owner, element)

    element.behavioredClassifier = parent
    return True


@ungroup.register(UML.BehavioredClassifier, UML.Behavior)
def behavior_ungroup(parent, element) -> bool:
    if element.behavioredClassifier is parent:
        del element.behavioredClassifier
        return True
    return False



import logging
from typing import Optional

from gaphas.decorators import g_async
from gaphas.handlemove import HandleMove
from gaphas.move import MoveType
from gaphas.view import GtkView
from gi.repository import GLib, Gtk

from gaphor.core.eventmanager import EventManager
from gaphor.diagram.diagramtoolbox import ItemFactory
from gaphor.diagram.event import DiagramItemPlaced
from gaphor.diagram.group import group
from gaphor.diagram.instanteditors import instant_editor
from gaphor.diagram.presentation import ElementPresentation
from gaphor.diagram.tools.dropzone import grow_parent

log = logging.getLogger(__name__)


class PlacementState:
    def __init__(
        self, factory: ItemFactory, event_manager: EventManager, handle_index: int
    ):
        self.factory = factory
        self.event_manager = event_manager
        self.handle_index = handle_index
        self.moving: Optional[MoveType] = None


def placement_tool(
    view: GtkView, factory: ItemFactory, event_manager, handle_index: int
):
    gesture = (
        Gtk.GestureDrag.new(view)
        if Gtk.get_major_version() == 3
        else Gtk.GestureDrag.new()
    )
    placement_state = PlacementState(factory, event_manager, handle_index)
    gesture.connect("drag-begin", on_drag_begin, placement_state)
    gesture.connect("drag-update", on_drag_update, placement_state)
    gesture.connect("drag-end", on_drag_end, placement_state)
    return gesture


def on_drag_begin(gesture, start_x, start_y, placement_state):
    view = gesture.get_widget()
    gesture.set_state(Gtk.EventSequenceState.CLAIMED)

    item = create_item(view, placement_state.factory, start_x, start_y)

    handle = item.handles()[placement_state.handle_index]
    if handle.movable:
        connect_opposite_handle(
            view, item, start_x, start_y, placement_state.handle_index
        )
        placement_state.moving = HandleMove(item, handle, view)
        placement_state.moving.start_move((start_x, start_y))
    else:
        placement_state.event_manager.handle(DiagramItemPlaced(item))

    view.selection.dropzone_item = None


def create_item(view, factory, x, y):
    selection = view.selection
    parent = selection.dropzone_item
    item = factory(view.model, parent)
    x, y = view.get_matrix_v2i(item).transform_point(x, y)
    item.matrix.translate(x, y)
    view.model.update_now({item})
    maybe_group(parent, item)
    selection.unselect_all()
    view.selection.focused_item = item
    return item


def maybe_group(parent, item):
    if parent and item.subject and group(parent.subject, item.subject):
        grow_parent(parent, item)
        item.change_parent(parent)


def connect_opposite_handle(view, new_item, x, y, handle_index):
    try:
        opposite = new_item.opposite(new_item.handles()[handle_index])
    except (KeyError, AttributeError):
        pass
    else:
        if opposite.connectable:
            vpos = (x, y)
            handle_move = HandleMove(new_item, opposite, view)
            if handle_move.glue(vpos):
                handle_move.connect(vpos)


def on_drag_update(gesture, offset_x, offset_y, placement_state):
    if placement_state.moving:
        _, x, y = gesture.get_start_point()
        placement_state.moving.move((x + offset_x, y + offset_y))


def on_drag_end(gesture, offset_x, offset_y, placement_state):
    if placement_state.moving:
        view = gesture.get_widget()
        _, x, y = gesture.get_start_point()
        item = placement_state.moving.item
        placement_state.moving.stop_move((x + offset_x, y + offset_y))
        connect_opposite_handle(view, item, x, y, placement_state.handle_index)
        placement_state.event_manager.handle(DiagramItemPlaced(item))
        open_editor(item, view, placement_state.event_manager)


@g_async(priority=GLib.PRIORITY_LOW)
def open_editor(item, view, event_manager):
    if isinstance(item, ElementPresentation):
        instant_editor(item, view, event_manager)



"""Actor item classes."""

from math import pi

from gaphas.constraint import constraint
from gaphas.geometry import distance_line_point
from gaphas.item import SE, SW
from gaphas.port import LinePort
from gaphas.position import Position
from gaphas.solver import REQUIRED, STRONG, variable

from gaphor import UML
from gaphor.diagram.presentation import Classified, ElementPresentation
from gaphor.diagram.shapes import Box, IconBox, Text, stroke
from gaphor.diagram.support import represents
from gaphor.diagram.text import FontStyle, FontWeight
from gaphor.UML.recipes import stereotypes_str

HEAD = 11
ARM = 19
NECK = 10
BODY = 20


@represents(UML.Actor)
class ActorItem(Classified, ElementPresentation):
    """Actor item is a classifier in icon mode.

    Maybe it should be possible to switch to compartment mode in the
    future.
    """

    text_height = variable(strength=REQUIRED, varname="_text_height")

    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=ARM * 2, height=HEAD + NECK + BODY + ARM)

        # Allow to connect to the bottom of the actor, below the actor name
        self._sub_text_port = LinePort(
            Position(0, 0, strength=STRONG), Position(0, 0, strength=STRONG)
        )
        self._ports.append(self._sub_text_port)

        self.text_height = 20.0

        add = diagram.connections.add_constraint
        add(
            self,
            constraint(
                horizontal=(self._handles[SW].pos, self._sub_text_port.start),
                delta=self.text_height,
            ),
        )
        add(
            self,
            constraint(vertical=(self._handles[SW].pos, self._sub_text_port.start)),
        )
        add(
            self,
            constraint(
                horizontal=(self._handles[SE].pos, self._sub_text_port.end),
                delta=self.text_height,
            ),
        )
        add(self, constraint(vertical=(self._handles[SE].pos, self._sub_text_port.end)))

        self.watch("subject[NamedElement].name")
        self.watch("subject.appliedStereotype.classifier.name")
        self.watch("subject[Classifier].isAbstract", self.update_shapes)

    def update_shapes(self, event=None):
        self.shape = IconBox(
            Box(
                draw=draw_actor,
            ),
            Text(
                text=lambda: stereotypes_str(self.subject),
            ),
            Text(
                text=lambda: self.subject.name or "",
                style={
                    "font-weight": FontWeight.BOLD,
                    "font-style": FontStyle.ITALIC
                    if self.subject and self.subject.isAbstract
                    else FontStyle.NORMAL,
                },
            ),
        )

    def update(self, context):
        super().update(context)
        _xs, ys = list(zip(*self.shape.sizes))
        self.text_height = sum(ys)

    def point(self, x: float, y: float) -> float:
        return min(  # type: ignore[no-any-return]
            super().point(x, y),
            distance_line_point(
                self._sub_text_port.start, self._sub_text_port.end, (x, y)
            )[0],
        )

    def postload(self):
        super().postload()
        self.diagram.connections.solve()


def draw_actor(box, context, bounding_box):
    """Draw actor's icon creature."""
    cr = context.cairo

    fx = bounding_box.width / (ARM * 2)
    fy = bounding_box.height / (HEAD + NECK + BODY + ARM)

    x = ARM * fx
    y = (HEAD / 2) * fy
    cy = HEAD * fy

    cr.move_to(x + HEAD * fy / 2.0, y)
    cr.arc(x, y, HEAD * fy / 2.0, 0, 2 * pi)

    cr.move_to(x, y + cy / 2)
    cr.line_to(ARM * fx, (HEAD + NECK + BODY) * fy)

    cr.move_to(0, (HEAD + NECK) * fy)
    cr.line_to(ARM * 2 * fx, (HEAD + NECK) * fy)

    cr.move_to(0, (HEAD + NECK + BODY + ARM) * fy)
    cr.line_to(ARM * fx, (HEAD + NECK + BODY) * fy)
    cr.line_to(ARM * 2 * fx, (HEAD + NECK + BODY + ARM) * fy)
    stroke(context, fill=False)



"""This module contains some support code for queries on lists.

Two mixin classes are provided:

1. ``querymixin``
2. ``recursemixin``

See the documentation on the mixins.
"""


from __future__ import annotations

import contextlib
from typing import Callable, Generic, Sequence, TypeVar, overload

__all__ = ["querymixin", "recursemixin"]


T = TypeVar("T")


def matcher(expr: str) -> Callable[[T], bool]:
    """Returns True if the expression returns True. The context for the
    expression is the element.

    Given a class:

    >>> class A:
    ...     def __init__(self, name):
    ...         self.name = name

    We can create a path for each object:

    >>> a = A("root")
    >>> a.a = A("level1")
    >>> a.b = A("b")
    >>> a.a.text = "help"

    If we want to match, ``it`` is used to refer to the subjected object:

    >>> matcher('it.name=="root"')(a)
    True
    >>> matcher('it.b.name=="b"')(a)
    True
    >>> matcher('it.name=="blah"')(a)
    False
    >>> matcher('it.nonexistent=="root"')(a)
    False

    NOTE: the object ``it`` was introduced since properties (descriptors) can
    not be executed from within a dictionary context.
    """

    compiled = compile(expr, "<matcher>", "eval")

    def real_matcher(element: T) -> bool:
        try:
            return bool(eval(compiled, {}, {"it": element}))
        except (AttributeError, NameError):
            # attribute does not (yet) exist
            return False

    return real_matcher


class querymixin:
    """Implementation of the matcher as a mixin for lists.

    Given a class:

    >>> class A:
    ...     def __init__(self, name):
    ...         self.name = name

    We can do nice things with this list:

    >>> class MList(querymixin, list):
    ...     pass
    >>> m = MList()
    >>> m.append(A("one"))
    >>> m.append(A("two"))
    >>> m.append(A("three"))
    >>> m[1].name
    'two'
    >>> m['it.name=="one"']  # doctest: +ELLIPSIS
    [<gaphor.core.modeling.listmixins.A object at 0x...>]
    >>> m['it.name=="two"', 0].name
    'two'
    """

    def __getitem__(self, key):
        try:
            # See if the list can deal with it (don't change default behaviour)
            return super().__getitem__(key)  # type: ignore[misc] # noqa: F821
        except TypeError:
            # Nope, try our matcher trick
            if isinstance(key, tuple):
                key, remainder = key[0], key[1:]
            else:
                remainder = None

            matched = list(filter(matcher(key), self))  # type: ignore[call-overload] # noqa: F821
            new_list = type(self)(matched)  # type: ignore[call-arg] # noqa: F821
            return new_list.__getitem__(*remainder) if remainder else new_list


def issafeiterable(obj):
    """Checks if the object is iterable, but not a string.

    >>> issafeiterable([])
    True
    >>> issafeiterable(set())
    True
    >>> issafeiterable({})
    True
    >>> issafeiterable(1)
    False
    >>> issafeiterable("text")
    False
    """
    with contextlib.suppress(TypeError):
        return iter(obj) and not isinstance(obj, str)
    return False


class recurseproxy(Generic[T]):
    """Proxy object (helper) for the recusemixin.

    The proxy has limited capabilities compared to a real list (or set):
    it can be iterated and a getitem can be performed. On the other
    side, the type of the original sequence is maintained, so getitem
    operations act as if they're executed on the original list.
    """

    def __init__(self, sequence: Sequence[T]):
        self.__sequence = sequence

    def __getitem__(self, key: int | slice) -> T:
        return self.__sequence.__getitem__(key)  # type: ignore[return-value]

    def __iter__(self):
        """Iterate over the items.

        If there is some level of nesting, the parent items are iterated
        as well.
        """
        return iter(self.__sequence)

    def __getattr__(self, key: str) -> recurseproxy[T]:
        """Create a new proxy for the attribute."""

        def mygetattr():
            sentinel = object()
            for e in self.__sequence:
                obj = getattr(e, key, sentinel)
                if obj is sentinel:
                    pass
                elif issafeiterable(obj):
                    yield from obj  # type: ignore[misc]
                else:
                    yield obj

        # Create a copy of the proxy type, including a copy of the sequence type
        return type(self)(type(self.__sequence)(mygetattr()))  # type: ignore[call-arg]


class recursemixin(Generic[T]):
    """Mixin class for lists, sets, etc. If data is requested using ``[:]``, a
    ``recurseproxy`` instance is created.

    The basic idea is to have a class that can contain children:

    >>> class A:
    ...     def __init__(self, name, *children):
    ...         self.name = name
    ...         self.children = list(children)
    ...
    ...     def dump(self, level=0):
    ...         print(" " * level, self.name)
    ...         for c in self.children:
    ...             c.dump(level + 1)

    Now if we make a (complex) structure out of it:

    >>> a = A("root", A("a", A("b"), A("c"), A("d")), A("e", A("one"), A("two")))
    >>> a.dump()  # doctest: +ELLIPSIS
     root
      a
       b
       c
       d
      e
       one
       two
    >>> a.children[1].name
    'e'

    Given ``a``, I want to iterate all grand-children (b, c, d, one, two) and
    the structure I want to do that with is:

      ``a.children[:].children``

    In order to do this we have to use a special list class, so we can handle
    our specific case. ``__getslice__`` should be overridden, so we can make it
    behave like a normal python object (legacy, yes...).

    >>> class rlist(recursemixin, list):
    ...     pass
    >>> class A:
    ...     def __init__(self, name, *children):
    ...         self.name = name
    ...         self.children = rlist(children)
    ...
    ...     def dump(self, level=0):
    ...         print(" " * level, self.name)
    ...         for c in self.children:
    ...             c.dump(level + 1)

    >>> a = A("root", A("a", A("b"), A("c"), A("d")), A("e", A("one"), A("two")))
    >>> a.children[1].name
    'e'

    Invoking ``a.children[:]`` should now return a recurseproxy object:

    >>> a.children[:]  # doctest: +ELLIPSIS
    <gaphor.core.modeling.listmixins.recurseproxy object at 0x...>
    >>> list(a.children[:].name)  # doctest: +ELLIPSIS
    ['a', 'e']

    Now calling a child on the list will return a list of all children:

    >>> a.children[:].children  # doctest: +ELLIPSIS
    <gaphor.core.modeling.listmixins.recurseproxy object at 0x...>
    >>> list(a.children[:].children)  # doctest: +ELLIPSIS
    [<gaphor.core.modeling.listmixins.A object at 0x...>, <gaphor.core.modeling.listmixins.A object at 0x...>, <gaphor.core.modeling.listmixins.A object at 0x...>, <gaphor.core.modeling.listmixins.A object at 0x...>, <gaphor.core.modeling.listmixins.A object at 0x...>]

    And of course we're interested in the names:

    >>> a.children[:].children.name  # doctest: +ELLIPSIS
    <gaphor.core.modeling.listmixins.recurseproxy object at 0x...>
    >>> list(a.children[:].children.name)
    ['b', 'c', 'd', 'one', 'two']
    """

    _recursemixin_trigger = slice(None, None, None)

    def proxy_class(self):
        return recurseproxy

    @overload
    def __getitem__(self, key: int) -> T:
        ...

    @overload
    def __getitem__(self, key: slice) -> recurseproxy[T]:
        ...

    def __getitem__(self, key):
        if key == self._recursemixin_trigger:
            return self.proxy_class()(self)
        else:
            return super().__getitem__(key)  # type: ignore[misc] # noqa: F821



"""About and help services.

(help browser anyone?)
"""

from gi.repository import Gtk

from gaphor.abc import ActionProvider, Service
from gaphor.application import distribution
from gaphor.core import action
from gaphor.i18n import translated_ui_string


def new_builder(ui_file):
    builder = Gtk.Builder()
    ui_file = f"{ui_file}.glade" if Gtk.get_major_version() == 3 else f"{ui_file}.ui"
    builder.add_from_string(translated_ui_string("gaphor.ui.help", ui_file))
    return builder


class HelpService(Service, ActionProvider):
    def __init__(self, main_window):
        self.main_window = main_window

    def shutdown(self):
        pass

    @property
    def window(self):
        return self.main_window.window

    @action(name="about")
    def about(self):
        builder = new_builder("about")
        about = builder.get_object("about")

        about.set_version(distribution().version)
        about.set_transient_for(self.window)

        if Gtk.get_major_version() == 3:
            about.run()
            about.destroy()
        else:
            about.set_modal(True)
            about.show()

    @action(name="win.shortcuts")
    def shortcuts(self):
        builder = Gtk.Builder()
        builder.add_from_string(translated_ui_string("gaphor.ui.help", "shortcuts.ui"))

        shortcuts = builder.get_object("shortcuts-gaphor")
        shortcuts.set_modal(True)
        shortcuts.set_transient_for(self.window)

        shortcuts.show()
        return shortcuts



from gaphor import UML
from gaphor.diagram.presentation import (
    AttachedPresentation,
    Classified,
    ElementPresentation,
    connect,
)
from gaphor.diagram.shapes import Box, JustifyContent, Text, TextAlign, draw_border
from gaphor.diagram.support import represents
from gaphor.diagram.text import FontStyle
from gaphor.UML.recipes import stereotypes_str


@represents(UML.Activity)
class ActivityItem(Classified, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=50, height=50)

        self.width = 100

        self.watch("subject[NamedElement].name").watch(
            "subject.appliedStereotype.classifier.name"
        ).watch("subject[Classifier].isAbstract", self.update_shapes).watch(
            "subject[Activity].node[ActivityParameterNode].parameter.name",
            self.update_parameters,
        ).watch(
            "subject[Activity].node[ActivityParameterNode].parameter.typeValue",
            self.update_parameters,
        )

    def postload(self):
        super().postload()
        self.update_parameters()

    def update_shapes(self, event=None):
        self.shape = Box(
            Text(
                text=lambda: stereotypes_str(self.subject),
                style={"text-align": TextAlign.LEFT},
            ),
            Text(
                text=lambda: self.subject.name or "",
                style={
                    "text-align": TextAlign.LEFT,
                    "font-style": FontStyle.ITALIC
                    if self.subject and self.subject.isAbstract
                    else FontStyle.NORMAL,
                },
            ),
            style={
                "padding": (4, 12, 4, 12),
                "border-radius": 20,
                "justify-content": JustifyContent.START,
            },
            draw=draw_border,
        )

    def update_parameters(self, event=None):
        diagram = self.diagram
        parameter_nodes = (
            [p for p in self.subject.node if isinstance(p, UML.ActivityParameterNode)]
            if self.subject
            else []
        )
        parameter_items = {
            i.subject: i
            for i in self.children
            if isinstance(i, ActivityParameterNodeItem)
        }

        for node in parameter_nodes:
            if node not in parameter_items:
                item = diagram.create(
                    ActivityParameterNodeItem, parent=self, subject=node
                )
                item.matrix.translate(0, 10)
                connect(item, item.handles()[0], self)

        for node in parameter_items:
            if node not in parameter_nodes:
                del self.children[parameter_items[node]]


class ActivityParameterNodeItem(AttachedPresentation[UML.ActivityParameterNode]):
    def __init__(self, diagram, id=None):
        super().__init__(
            diagram,
            id,
            shape=Box(
                Text(
                    text=lambda: self.subject.parameter.name or "",
                ),
                style={"padding": (4, 12, 4, 12), "background-color": (1, 1, 1, 1)},
                draw=draw_border,
            ),
            width=100,
            height=30,
        )

        self.watch("subject[ActivityParameterNode].parameter.name")



from gaphor.C4Model.c4model import C4Container, C4Database
from gaphor.diagram.iconname import get_default_icon_name, get_icon_name


@get_icon_name.register(C4Container)
def get_name_for_class(element):
    if element.type == "Software System":
        return "gaphor-c4-software-system-symbolic"
    elif element.type == "Component":
        return "gaphor-c4-component-symbolic"
    else:
        return get_default_icon_name(element)


@get_icon_name.register(C4Database)
def get_database_name(element):
    return get_default_icon_name(element)



from gaphor import UML
from gaphor.diagram.propertypages import ComboModel, PropertyPageBase, PropertyPages
from gaphor.i18n import gettext
from gaphor.transaction import transactional
from gaphor.UML.classes.classespropertypages import new_builder
from gaphor.UML.classes.dependency import DependencyItem


@PropertyPages.register(DependencyItem)
class DependencyPropertyPage(PropertyPageBase):
    """Dependency item editor."""

    order = 20

    DEPENDENCY_TYPES = (
        (gettext("Dependency"), UML.Dependency),
        (gettext("Usage"), UML.Usage),
        (gettext("Realization"), UML.Realization),
        (gettext("Implementation"), UML.InterfaceRealization),
    )

    def __init__(self, item):
        super().__init__()
        self.item = item
        self.watcher = self.item.watcher()
        self.builder = new_builder(
            "dependency-editor",
            signals={
                "dependency-type-changed": (self._on_dependency_type_change,),
                "automatic-changed": (self._on_auto_dependency_change,),
                "dependency-type-destroy": (self.watcher.unsubscribe_all,),
            },
        )

    def construct(self):
        dependency_combo = self.builder.get_object("dependency-combo")
        model = ComboModel(self.DEPENDENCY_TYPES)
        dependency_combo.set_model(model)

        automatic = self.builder.get_object("automatic")
        automatic.set_active(self.item.auto_dependency)

        self.update()

        self.watcher.watch("subject", self._on_subject_change)

        return self.builder.get_object("dependency-editor")

    def _on_subject_change(self, event):
        self.update()

    def update(self):
        """Update dependency type combo box.

        Disallow dependency type when dependency is established.
        """
        combo = self.builder.get_object("dependency-combo")
        if combo.get_model():
            item = self.item
            index = combo.get_model().get_index(item.dependency_type)
            combo.props.sensitive = not item.auto_dependency
            combo.set_active(index)

    @transactional
    def _on_dependency_type_change(self, combo):
        cls = combo.get_model().get_value(combo.get_active())
        self.item.dependency_type = cls
        if subject := self.item.subject:
            UML.recipes.swap_element(subject, cls)
            self.item.request_update()

    @transactional
    def _on_auto_dependency_change(self, switch, gparam):
        self.item.auto_dependency = switch.get_active()
        self.update()



"""The file service is responsible for loading and saving the user data."""

from __future__ import annotations

import logging
from pathlib import Path
from queue import Queue

from gaphas.decorators import g_async
from gi.repository import Gtk

if Gtk.get_major_version() == 4:
    from gi.repository import Adw

from gaphor import UML
from gaphor.abc import ActionProvider, Service
from gaphor.core import action, event_handler, gettext
from gaphor.core.modeling import Diagram, StyleSheet
from gaphor.event import (
    ModelLoaded,
    ModelSaved,
    SessionCreated,
    SessionShutdown,
    SessionShutdownRequested,
)
from gaphor.storage import storage
from gaphor.storage.parser import MergeConflictDetected
from gaphor.ui.errorhandler import error_handler
from gaphor.ui.filedialog import GAPHOR_FILTER, save_file_dialog
from gaphor.ui.statuswindow import StatusWindow

DEFAULT_EXT = ".gaphor"
MAX_RECENT = 10

log = logging.getLogger(__name__)


def error_message(e):
    if not isinstance(e, IOError):
        return gettext(
            "Gaphor was not able to store the model, probably due to an internal error:\n{exc}\nIf you think this is a bug, please contact the developers."
        ).format(exc=str(e))
    if e.errno == 13:
        return gettext(
            "You do not have the permissions necessary to save the model.\nPlease check that you typed the location correctly and try again."
        )
    elif e.errno == 28:
        return gettext(
            "You do not have enough free space on the device to save the model.\nPlease free up some disk space and try again or save it in a different location."
        )
    return gettext(
        "The model cannot be stored at this location:\n{exc}\nPlease check that you typed the location correctly and try again."
    ).format(exc=str(e))


def load_default_model(element_factory):
    element_factory.flush()
    with element_factory.block_events():
        element_factory.create(StyleSheet)
        model = element_factory.create(UML.Package)
        model.name = gettext("New model")
        diagram = element_factory.create(Diagram)
        diagram.element = model
        diagram.name = gettext("main")
    element_factory.model_ready()


class FileManager(Service, ActionProvider):
    """The file service, responsible for loading and saving Gaphor models."""

    def __init__(self, event_manager, element_factory, modeling_language, main_window):
        """File manager constructor.

        There is no current filename yet.
        """
        self.event_manager = event_manager
        self.element_factory = element_factory
        self.modeling_language = modeling_language
        self.main_window = main_window
        self._filename = None

        event_manager.subscribe(self._on_session_shutdown_request)
        event_manager.subscribe(self._on_session_created)

    def shutdown(self):
        """Called when shutting down the file manager service."""
        self.event_manager.unsubscribe(self._on_session_shutdown_request)
        self.event_manager.unsubscribe(self._on_session_created)

    @property
    def filename(self):
        """Return the current file name.

        This method is used by the filename property.
        """
        return self._filename

    @filename.setter
    def filename(self, filename):
        """Sets the current file name.

        This method is used by the filename property. Setting the
        current filename will update the recent file list.
        """

        if filename != self._filename:
            self._filename = Path(filename) if filename else None

    def load(self, filename):
        """Load the Gaphor model from the supplied file name.

        A status window displays the loading progress.  The load
        generator updates the progress queue.  The loader is passed to a
        GIdleThread which executes the load generator.  If loading is
        successful, the filename is set.
        """
        # First claim file name, so any other files will be opened in a different session
        self.filename = Path(filename)
        queue: Queue[int] = Queue(0)
        status_window = None

        @g_async()
        def create_status_window():
            nonlocal status_window
            status_window = StatusWindow(
                gettext("Loading…"),
                gettext("Loading model from {filename}").format(filename=filename),
                parent=self.main_window.window if self.main_window else None,
                queue=queue,
            )

        def done():
            if status_window:
                status_window.destroy()

        create_status_window()
        self._load_async(filename, queue, done)

    def load_template(self, template):
        storage.load(template, self.element_factory, self.modeling_language)
        self.event_manager.handle(ModelLoaded(self))

    def _load_async(self, filename: Path, queue=None, done=None):
        assert isinstance(filename, Path)

        # Use low prio, so screen updates do happen
        @g_async()
        def async_loader():
            try:
                try:
                    yield from open_model(encoding="utf-8")
                except UnicodeDecodeError:
                    # try to load without encoding, for older models saved on windows
                    yield from open_model(encoding=None)

                self.event_manager.handle(ModelLoaded(self, filename))
            except MergeConflictDetected:
                self.filename = None
                error_handler(
                    message=gettext("Merge conflict in model “{filename}”.").format(
                        filename=filename
                    ),
                    secondary_message=gettext(
                        "It looks like this model file contains a merge conflict."
                    ),
                    close=lambda: self.event_manager.handle(SessionShutdown(self)),
                )
                # For now load the default model until we allow users to resolve the merge conflict.
                load_default_model(self.element_factory)
            except Exception:
                self.filename = None
                error_handler(
                    message=gettext("Unable to open model “{filename}”.").format(
                        filename=filename
                    ),
                    secondary_message=gettext(
                        "This file does not contain a valid Gaphor model."
                    ),
                    window=self.main_window.window,
                    close=lambda: self.event_manager.handle(SessionShutdown(self)),
                )
            finally:
                if done:
                    done()

        def open_model(encoding):
            with open(filename, encoding=encoding) as file_obj:
                for percentage in storage.load_generator(
                    file_obj,
                    self.element_factory,
                    self.modeling_language,
                ):
                    if queue:
                        queue.put(percentage)
                    yield percentage

        for _ in async_loader():
            pass

    def save(self, filename, on_save_done=None):
        """Save the current UML model to the specified file name.

        Before writing the model file, this will verify that there are
        no orphan references.  It will also verify that the filename has
        the correct extension.  A status window is displayed while the
        GIdleThread is executed.  This thread actually saves the model.
        """

        if not filename or (filename.exists() and not filename.is_file()):
            return

        main_window = self.main_window
        queue: Queue[int] = Queue()
        status_window = StatusWindow(
            gettext("Saving…"),
            gettext("Saving model to {filename}").format(filename=filename),
            parent=main_window.window if main_window else None,
            queue=queue,
        )

        @g_async()
        def async_saver():
            try:
                with filename.open("w", encoding="utf-8") as out:
                    for percentage in storage.save_generator(out, self.element_factory):
                        queue.put(percentage)
                        yield

                self.filename = filename
                self.event_manager.handle(ModelSaved(self, filename))
            except Exception as e:
                error_handler(
                    message=gettext("Unable to save model “{filename}”.").format(
                        filename=filename
                    ),
                    secondary_message=error_message(e),
                    window=main_window.window if main_window else None,
                )
                raise
            finally:
                status_window.destroy()
            if on_save_done:
                on_save_done()

        for _ in async_saver():
            pass

    @action(name="file-save", shortcut="<Primary>s")
    def action_save(self):
        """Save the file. Depending on if there is a file name, either perform
        the save directly or present the user with a save dialog box.

        Returns True if the saving actually succeeded.
        """

        if filename := self.filename:
            self.save(filename)
        else:
            self.action_save_as()

    @action(name="file-save-as", shortcut="<Primary><Shift>s")
    def action_save_as(self):
        """Save the model in the element_factory by allowing the user to select
        a file name.

        Returns True if the saving actually happened.
        """

        save_file_dialog(
            gettext("Save Gaphor Model As"),
            self.save,
            parent=self.main_window.window,
            filename=self.filename,
            extension=".gaphor",
            filters=GAPHOR_FILTER,
        )

    @event_handler(SessionCreated)
    def _on_session_created(self, event: SessionCreated) -> None:
        if event.filename:
            self.load(event.filename)
        elif event.template:
            self.load_template(event.template)
        else:
            load_default_model(self.element_factory)

    @event_handler(SessionShutdownRequested)
    def _on_session_shutdown_request(self, event: SessionShutdownRequested) -> None:
        """Ask user to close window if the model has changed.

        The user is asked to either discard the changes, keep the
        application running or save the model and quit afterwards.
        """

        def confirm_shutdown():
            self.event_manager.handle(SessionShutdown(self))

        def response(answer):
            # Gtk.ResponseType.YES is GTK3, save is GTK4
            if answer in [Gtk.ResponseType.YES, "save"]:
                if filename := self.filename:
                    self.save(filename, on_save_done=confirm_shutdown)

                else:
                    save_file_dialog(
                        gettext("Save Gaphor Model As"),
                        lambda filename: self.save(
                            filename, on_save_done=confirm_shutdown
                        ),
                        parent=self.main_window.window,
                        filename=self.filename,
                        extension=".gaphor",
                        filters=GAPHOR_FILTER,
                    )
            # Gtk.ResponseType.REJECT is GTK3, discard is GTK4
            elif answer in [Gtk.ResponseType.REJECT, "discard"]:
                confirm_shutdown()

        if self.main_window.model_changed:
            save_changes_before_closing_dialog(self.main_window.window, response)
        else:
            confirm_shutdown()


def save_changes_before_closing_dialog(window: Gtk.Window, handler) -> None:
    title = gettext("Save Changes?")
    body = gettext(
        "The open model contains unsaved changes. Changes which are not saved will be permanently lost."
    )
    if Gtk.get_major_version() == 3:
        dialog = Gtk.MessageDialog(
            message_type=Gtk.MessageType.WARNING,
            text=title,
            secondary_text=body,
        )
        dialog.set_transient_for(window)
        dialog.set_modal(True)
        dialog.add_buttons(
            gettext("Cancel"),
            Gtk.ResponseType.CANCEL,
            gettext("Discard"),
            Gtk.ResponseType.REJECT,
            gettext("Save"),
            Gtk.ResponseType.YES,
        )
        dialog.set_default_response(Gtk.ResponseType.YES)
    else:
        dialog = Adw.MessageDialog.new(
            window,
            title,
        )
        dialog.set_body(body)
        dialog.add_response("cancel", gettext("Cancel"))
        dialog.add_response("discard", gettext("Discard"))
        dialog.add_response("save", gettext("Save"))
        dialog.set_response_appearance("discard", Adw.ResponseAppearance.DESTRUCTIVE)
        dialog.set_response_appearance("save", Adw.ResponseAppearance.SUGGESTED)
        dialog.set_default_response("save")
        dialog.set_close_response("cancel")

    def response(dialog, answer):
        dialog.destroy()
        handler(answer)

    dialog.connect("response", response)
    dialog.show()



"""The definition for the actions section of the toolbox."""
from functools import partial

from gaphas.item import SE

from gaphor import UML
from gaphor.core import gettext
from gaphor.diagram.diagramtoolbox import ToolDef, ToolSection, new_item_factory
from gaphor.UML import diagramitems
from gaphor.UML.recipes import owner_package
from gaphor.UML.toolboxconfig import namespace_config


def activity_config(new_item, name=None):
    subject = new_item.subject
    if name:
        subject.name = gettext("New {name}").format(name=name)
    if subject.activity:
        return

    diagram = new_item.diagram

    if owner_activity := UML.recipes.owner_of_type(diagram, UML.Activity):
        subject.activity = owner_activity
        return

    package = owner_package(diagram.owner)

    activities = (
        [i for i in package.ownedType if isinstance(i, UML.Activity)]
        if package
        else diagram.model.lselect(
            lambda e: isinstance(e, UML.Activity) and e.package is None
        )
    )
    if activities:
        subject.activity = activities[0]
    else:
        activity = subject.model.create(UML.Activity)
        activity.name = gettext("Activity")
        activity.package = package
        subject.activity = activity


def partition_config(new_item):
    activity_config(new_item)
    subject = new_item.subject
    subject.name = gettext("Swimlane One")
    new_item.partition = subject

    partition = subject.model.create(UML.ActivityPartition)
    partition.name = gettext("Swimlane Two")
    partition.activity = subject.activity
    new_item.partition = partition


actions = ToolSection(
    gettext("Actions"),
    (
        ToolDef(
            "toolbox-activity",
            gettext("Activity"),
            "gaphor-activity-symbolic",
            None,
            new_item_factory(
                diagramitems.ActivityItem,
                UML.Activity,
                config_func=namespace_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-action",
            gettext("Action"),
            "gaphor-action-symbolic",
            "a",
            new_item_factory(
                diagramitems.ActionItem,
                UML.Action,
                config_func=partial(activity_config, name=gettext("Action")),
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-initial-node",
            gettext("Initial node"),
            "gaphor-initial-node-symbolic",
            "j",
            new_item_factory(
                diagramitems.InitialNodeItem,
                UML.InitialNode,
                config_func=activity_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-activity-final-node",
            gettext("Activity final node"),
            "gaphor-activity-final-node-symbolic",
            "f",
            new_item_factory(
                diagramitems.ActivityFinalNodeItem,
                UML.ActivityFinalNode,
                config_func=activity_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-flow-final-node",
            gettext("Flow final node"),
            "gaphor-flow-final-node-symbolic",
            "w",
            new_item_factory(
                diagramitems.FlowFinalNodeItem,
                UML.FlowFinalNode,
                config_func=activity_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-decision-node",
            gettext("Decision/merge node"),
            "gaphor-decision-node-symbolic",
            "g",
            new_item_factory(
                diagramitems.DecisionNodeItem,
                UML.DecisionNode,
                config_func=activity_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-fork-node",
            gettext("Fork/join node"),
            "gaphor-fork-node-symbolic",
            "<Shift>R",
            new_item_factory(
                diagramitems.ForkNodeItem,
                UML.JoinNode,
                config_func=activity_config,
            ),
            handle_index=1,
        ),
        ToolDef(
            "toolbox-object-node",
            gettext("Object node"),
            "gaphor-object-node-symbolic",
            "<Shift>O",
            new_item_factory(
                diagramitems.ObjectNodeItem,
                UML.ObjectNode,
                config_func=partial(activity_config, name=gettext("Object node")),
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-partition",
            gettext("Swimlane"),
            "gaphor-activity-partition-symbolic",
            "<Shift>P",
            new_item_factory(
                diagramitems.PartitionItem,
                UML.ActivityPartition,
                config_func=partition_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-control-flow",
            gettext("Control flow"),
            "gaphor-control-flow-symbolic",
            "<Shift>F",
            new_item_factory(diagramitems.ControlFlowItem),
        ),
        ToolDef(
            "toolbox-object-flow",
            gettext("Object flow"),
            "gaphor-object-flow-symbolic",
            "<Shift>Y",
            new_item_factory(diagramitems.ObjectFlowItem),
        ),
        ToolDef(
            "toolbox-send-signal-action",
            gettext("Send signal action"),
            "gaphor-send-signal-action-symbolic",
            None,
            new_item_factory(
                diagramitems.SendSignalActionItem,
                UML.SendSignalAction,
                config_func=partial(activity_config, name=gettext("Send signal")),
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-accept-event-action",
            gettext("Accept event action"),
            "gaphor-accept-event-action-symbolic",
            None,
            new_item_factory(
                diagramitems.AcceptEventActionItem,
                UML.AcceptEventAction,
                config_func=partial(activity_config, name=gettext("Accept event")),
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-input-pin",
            gettext("Input pin"),
            "gaphor-input-pin-symbolic",
            None,
            new_item_factory(diagramitems.InputPinItem),
        ),
        ToolDef(
            "toolbox-output-pin",
            gettext("Output pin"),
            "gaphor-output-pin-symbolic",
            None,
            new_item_factory(diagramitems.OutputPinItem),
        ),
    ),
)



#!/usr/bin/env python

#  GTK Interactive Console
#  (C) 2003, Jon Anderson
#  See www.python.org/2.2/license.html for
#  license details.
#
#  Took some good ideas from the GEdit Python Console:
#  https://gitlab.gnome.org/GNOME/gedit/-/blob/master/plugins/pythonconsole/pythonconsole/console.py
#  Licensed under GPL


import code
import sys
import textwrap
import traceback
from typing import Dict, List

import jedi
from gi.repository import Gdk, GLib, Gtk, Pango

from gaphor.i18n import gettext

banner = gettext(
    """\
Gaphor Interactive Python Console
{version}
Type "help" for more information.
"""
).format(version=sys.version)


def _text_tag(name, **properties):
    tag = Gtk.TextTag.new(name)
    for prop, val in properties.items():
        tag.set_property(prop, val)
    return name, tag


prompt = {
    "ps1": ">>> ",
    "ps2": "... ",
}

style = dict(
    [
        _text_tag("banner", foreground="#77767b"),
        _text_tag("ps1", editable=False),
        _text_tag("ps2", foreground="#77767b", editable=False),
        _text_tag("stdout", editable=False),
        _text_tag(
            "stderr", style=Pango.Style.ITALIC, foreground="#cc0000", editable=False
        ),
    ]
)


def docstring_dedent(docstr: str) -> str:
    if docstr.startswith(" "):
        return textwrap.dedent(docstr)

    first_line, remaining_text = docstr.split("\n", 1)

    return "\n".join([first_line, textwrap.dedent(remaining_text)])


class Help:
    def __init__(self, writer, locals=None):
        if locals is None:
            locals = {}
        self._writer = writer
        self._locals = locals

    def __call__(self, obj=None):
        helptext = docstring_dedent(obj.__doc__ if obj else str(self))
        with self._writer as writer:
            writer.write("\n")
            writer.write(textwrap.indent(helptext, prefix="  "))
            writer.write("\n" if helptext.endswith("\n") else "\n\n")

    def __str__(self):
        intro = textwrap.dedent(
            """\

        Usage: help(object)

        The following functions/variables are defined:
        """
        )

        members = "\n".join(
            f"- {key}" for key in self._locals.keys() if not key.startswith("_")
        )

        return textwrap.indent(intro + members + "\n", prefix="  ")

    def __repr__(self):
        return str(self)


class TextViewWriter:
    """A Multiplexing output stream.

    It can replace another stream, and tee output to the original stream
    and to a GTK textview.
    """

    def __init__(self, name, view):
        self.name = name
        self.out = getattr(sys, name)
        self.view = view

    def write(self, text):
        buffer = self.view.get_buffer()
        end = buffer.get_end_iter()
        buffer.insert_with_tags(end, text, style[self.name])

    def __enter__(self):
        setattr(sys, self.name, self)
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        setattr(sys, self.name, self.out)


class ExceptionWriter:
    """A Multiplexing output stream.

    It can replace another stream, and tee output to the original stream
    and to a GTK textview.
    """

    def excepthook(self, exc_type, exc_value, exc_traceback):
        traceback.print_exception(
            exc_type, value=exc_value, tb=exc_traceback, file=sys.stderr
        )

    def __enter__(self):
        self.orig_excepthook = sys.excepthook
        sys.excepthook = self.excepthook
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        sys.excepthook = self.orig_excepthook


class GTKInterpreterConsole(Gtk.ScrolledWindow):
    """An Interactive Console for GTK.

    It's an actual widget, so it can be dropped in just about anywhere.
    """

    __gtype_name__ = "GTKInterpreterConsole"

    def __init__(self, locals: Dict[str, object], banner=banner):
        Gtk.ScrolledWindow.__init__(self)

        self.set_vexpand(True)
        self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)

        self.text = Gtk.TextView()
        self.text.set_wrap_mode(True)
        self.text.set_monospace(True)

        self.interpreter = code.InteractiveInterpreter(locals)
        self.locals = locals

        self.buffer: List[str] = []
        self.history: List[str] = []
        self.banner = banner

        if Gtk.get_major_version() == 3:
            self.text_controller = Gtk.EventControllerKey.new(self.text)
        else:
            self.text_controller = Gtk.EventControllerKey.new()
            self.text.add_controller(self.text_controller)

        self.text_controller.connect("key-pressed", self.key_pressed)

        self.current_history = -1

        buffer = self.text.get_buffer()
        buffer.create_mark("input", buffer.get_end_iter(), True)

        for tag in style.values():
            buffer.get_tag_table().add(tag)

        self.stdout = TextViewWriter("stdout", self.text)
        self.stderr = TextViewWriter("stderr", self.text)
        self.excepthook = ExceptionWriter()

        self.current_prompt = "ps1"
        locals["help"] = Help(self.stdout, locals)

        self.write(self.banner, style["banner"])
        self.prompt()

        if Gtk.get_major_version() == 3:
            self.add(self.text)
            self.text.show()
        else:
            self.set_child(self.text)

    def reset_buffer(self):
        self.buffer = []

    def prompt(self):
        self.write(prompt[self.current_prompt], style[self.current_prompt])
        self.move_input_mark()
        GLib.idle_add(self.scroll_to_end)

    def write(self, text, style=None):
        buffer = self.text.get_buffer()
        start, end = buffer.get_bounds()
        if style:
            buffer.insert_with_tags(end, text, style)
        else:
            buffer.insert(end, text)

        buffer.place_cursor(buffer.get_end_iter())

    def move_input_mark(self):
        buffer = self.text.get_buffer()
        input_mark = buffer.get_mark("input")
        buffer.move_mark(input_mark, buffer.get_end_iter())

    def scroll_to_end(self):
        buffer = self.text.get_buffer()
        input_mark = buffer.get_mark("input")
        self.text.scroll_to_mark(input_mark, 0, True, 1, 1)
        return GLib.SOURCE_REMOVE

    def push(self, line):
        self.buffer.append(line)
        if len(line) > 0:
            self.history.append(line)

        source = "\n".join(self.buffer)

        with self.stdout, self.stderr, self.excepthook:
            more = self.interpreter.runsource(source, "<<console>>")

        if not more:
            self.reset_buffer()

        return more

    def key_pressed(self, widget, keyval, keycode, state):
        if keyval == Gdk.keyval_from_name("Return"):
            return self.execute_line()
        if keyval == Gdk.keyval_from_name("Tab"):
            return self.complete_line()

        if keyval == Gdk.keyval_from_name("Up"):
            self.current_history = self.current_history - 1
            self.current_history = max(self.current_history, -len(self.history))
            return self.show_history()
        if keyval == Gdk.keyval_from_name("Down"):
            self.current_history = self.current_history + 1
            self.current_history = min(self.current_history, 0)
            return self.show_history()

        ctrl = state & Gdk.ModifierType.CONTROL_MASK
        if (keyval == Gdk.keyval_from_name("Home")) or (ctrl and keyval == Gdk.KEY_a):
            line_count = self.text.get_buffer().get_line_count() - 1
            start = self.text.get_buffer().get_iter_at_line_offset(line_count, 4)
            self.text.get_buffer().place_cursor(start)
            return True

        if ctrl and keyval == Gdk.KEY_e:
            end = self.text.get_buffer().get_end_iter()
            self.text.get_buffer().place_cursor(end)
            return True

        if (
            ctrl
            and keyval == Gdk.KEY_c
            and not self.text.get_buffer().get_selection_bounds()
        ):
            self.write("^C\n")
            self.reset_buffer()
            self.current_prompt = "ps1"
            self.prompt()
            return True

        return False

    def show_history(self):
        if self.current_history != 0:
            self.replace_line(self.history[self.current_history])

        return True

    def current_line(self):
        start, end = self.current_line_bounds()
        return self.text.get_buffer().get_text(start, end, True)

    def current_line_bounds(self):
        buffer = self.text.get_buffer()
        input_mark = buffer.get_mark("input")
        start = buffer.get_iter_at_mark(input_mark)
        end = buffer.get_end_iter()
        return start, end

    def replace_line(self, txt):
        start, end = self.current_line_bounds()
        self.text.get_buffer().delete(start, end)
        self.write(txt)

    def execute_line(self):
        line = self.current_line()

        self.write("\n")

        more = self.push(line)

        self.current_prompt = "ps2" if more else "ps1"
        self.prompt()
        self.current_history = 0

        return True

    def complete_line(self):
        line = self.current_line()
        source = "\n".join(self.buffer + [line])
        script = jedi.Interpreter(source, [self.locals])
        completions = script.complete()

        if len(completions) > 1:
            max_len = max(map(lambda c: len(c.name), completions)) + 2
            per_line = 76 // max_len
            for i, c in enumerate(completions):
                if i % per_line == 0:
                    self.write("\n")
                self.write(c.name, style["ps1"])
                self.write(" " * (max_len - len(c.name)), style["ps1"])
            self.write("\n")
            self.prompt()
            self.write(line)
        elif len(completions) == 1:
            self.write(completions[0].complete)

        return True


def main(main_loop=True):
    window = Gtk.ApplicationWindow()

    window.set_default_size(640, 480)
    console = GTKInterpreterConsole(locals())

    def key_event(widget, keyval, keycode, state):
        if keyval == Gdk.KEY_d and state & Gdk.ModifierType.CONTROL_MASK:
            app.quit()
        return False

    console.text_controller.connect("key-pressed", key_event)

    if Gtk.get_major_version() == 3:
        window.connect("destroy", lambda w: app.quit())
        window.add(console)
        window.show_all()
    else:
        window.set_child(console)
        window.show()

    if main_loop:

        def on_activate(app):
            app.add_window(window)

        app = Gtk.Application(application_id="org.gaphor.Console")
        app.connect("activate", on_activate)
        app.run()


if __name__ == "__main__":
    main()



"""
ExtensionItem -- Graphical representation of an association.
"""

from gaphor import UML
from gaphor.diagram.presentation import LinePresentation, Named
from gaphor.diagram.shapes import Box, Text
from gaphor.diagram.support import represents
from gaphor.UML.recipes import stereotypes_str


@represents(UML.Extension)
class ExtensionItem(Named, LinePresentation):
    """ExtensionItem represents associations.

    An ExtensionItem has two ExtensionEnd items. Each ExtensionEnd item
    represents a Property (with Property.association == my association).
    """

    def __init__(self, diagram, id=None):
        super().__init__(
            diagram,
            id,
            shape_middle=Box(
                Text(
                    text=lambda: stereotypes_str(self.subject),
                ),
                Text(text=lambda: self.subject.name or ""),
            ),
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject.appliedStereotype.classifier.name")

    def draw_head(self, context):
        cr = context.cairo
        cr.move_to(0, 0)
        cr.line_to(15, -10)
        cr.line_to(15, 10)
        cr.line_to(0, 0)
        cr.fill_preserve()
        cr.move_to(15, 0)



"""The action definition for the C4 Model toolbox."""
from functools import partial

from gaphas.item import SE

from gaphor.C4Model import c4model, diagramitems
from gaphor.core import gettext
from gaphor.diagram.diagramtoolbox import (
    DiagramType,
    DiagramTypes,
    ToolboxDefinition,
    ToolDef,
    ToolSection,
    general_tools,
    new_item_factory,
)
from gaphor.UML import diagramitems as uml_items
from gaphor.UML.actions.actionstoolbox import actions
from gaphor.UML.classes.classestoolbox import classes
from gaphor.UML.interactions.interactionstoolbox import interactions
from gaphor.UML.states.statestoolbox import states
from gaphor.UML.toolboxconfig import default_namespace, namespace_config


def software_system_config(new_item):
    default_namespace(new_item)
    subject = new_item.subject
    subject.type = "Software System"
    subject.name = gettext("New Software System")


def container_config(new_item):
    default_namespace(new_item)
    subject = new_item.subject
    subject.type = "Container"
    subject.name = gettext("New Container")


def container_database_config(new_item):
    default_namespace(new_item)
    subject = new_item.subject
    subject.type = "Container"
    subject.technology = gettext("Database")
    subject.name = gettext("New Database")


def component_config(new_item):
    default_namespace(new_item)
    subject = new_item.subject
    subject.type = "Component"
    subject.name = gettext("New Component")


c4 = ToolSection(
    gettext("C4 Model"),
    (
        ToolDef(
            "c4-person",
            gettext("Person"),
            "gaphor-c4-person-symbolic",
            "P",
            new_item_factory(
                diagramitems.C4PersonItem,
                c4model.C4Person,
                config_func=partial(namespace_config, name=gettext("Person")),
            ),
            handle_index=SE,
        ),
        ToolDef(
            "c4-software-system",
            gettext("Software System"),
            "gaphor-c4-software-system-symbolic",
            "<Shift>S",
            new_item_factory(
                diagramitems.C4ContainerItem,
                c4model.C4Container,
                config_func=software_system_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "c4-container",
            gettext("Container"),
            "gaphor-c4-container-symbolic",
            "u",
            new_item_factory(
                diagramitems.C4ContainerItem,
                c4model.C4Container,
                config_func=container_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "c4-container-database",
            gettext("Container: Database"),
            "gaphor-c4-database-symbolic",
            "<Shift>B",
            new_item_factory(
                diagramitems.C4DatabaseItem,
                c4model.C4Database,
                config_func=container_database_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "c4-component",
            gettext("Component"),
            "gaphor-c4-component-symbolic",
            "<Shift>X",
            new_item_factory(
                diagramitems.C4ContainerItem,
                c4model.C4Container,
                config_func=component_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "c4-dependency",
            gettext("Dependency"),
            "gaphor-dependency-symbolic",
            "d",
            new_item_factory(uml_items.DependencyItem),
            handle_index=0,
        ),
    ),
)


c4model_toolbox_actions: ToolboxDefinition = (
    general_tools,
    c4,
    classes,
    actions,
    interactions,
    states,
)

c4model_diagram_types: DiagramTypes = (
    DiagramType("c4", gettext("New C4 Diagram"), (c4,)),
    DiagramType("cls", gettext("New Class Diagram"), (classes,)),
    DiagramType("act", gettext("New Activity Diagram"), (actions,)),
    DiagramType("sd", gettext("New Sequence Diagram"), (interactions,)),
    DiagramType("stm", gettext("New State Machine Diagram"), (states,)),
)



import logging
import os.path

from gi.repository import Gtk

if Gtk.get_major_version() == 4:
    from gi.repository import Adw

from pathlib import Path

from gaphor.abc import ActionProvider, Service
from gaphor.core import action, gettext
from gaphor.ui.filedialog import open_file_dialog

log = logging.getLogger(__name__)


FILTERS = [
    (gettext("All Gaphor Models"), "*.gaphor", "application/x-gaphor"),
]


class AppFileManager(Service, ActionProvider):
    """Handle application level file loading."""

    def __init__(self, application):
        self.application = application
        self.last_dir = None

    def shutdown(self):
        pass

    @property
    def parent_window(self):
        return (
            self.application.active_session.get_service("main_window").window
            if self.application.active_session
            else None
        )

    @action(name="app.file-open")
    def action_open(self):
        """This menu action opens the standard model open dialog."""

        def open_files(filenames):
            for filename in filenames:
                if self.application.has_session(filename):
                    name = Path(filename).name
                    title = gettext("Switch to {name}?").format(name=name)
                    body = gettext(
                        "{name} is already opened. Do you want to switch to the opened window instead?"
                    ).format(name=name)
                    if Gtk.get_major_version() == 3:
                        dialog = Gtk.MessageDialog(
                            message_type=Gtk.MessageType.QUESTION,
                            buttons=Gtk.ButtonsType.YES_NO,
                            text=title,
                            secondary_text=body,
                        )
                        dialog.set_transient_for(self.parent_window)

                        dialog.set_modal(True)
                    else:
                        dialog = Adw.MessageDialog.new(
                            self.parent_window,
                            title,
                        )
                        dialog.set_body(body)
                        dialog.add_response("open", gettext("Open Again"))
                        dialog.add_response("switch", gettext("Switch"))
                        dialog.set_response_appearance(
                            "switch", Adw.ResponseAppearance.SUGGESTED
                        )
                        dialog.set_default_response("switch")
                        dialog.set_close_response("open")

                    def response(dialog, answer):
                        # Gtk.ResponseType.NO is for GTK3, open is for GTK4
                        force_new_session = answer in [Gtk.ResponseType.NO, "open"]
                        dialog.destroy()
                        self.application.new_session(
                            filename=filename, force=force_new_session
                        )

                    dialog.connect("response", response)
                    dialog.show()
                else:
                    self.application.new_session(filename=filename)

                self.last_dir = os.path.dirname(filename)

        open_file_dialog(
            gettext("Open a Model"),
            open_files,
            parent=self.parent_window,
            dirname=self.last_dir,
            filters=FILTERS,
        )



"""Application lifecycle events are managed here."""

from __future__ import annotations

from pathlib import Path

from gaphor.abc import Service


class ServiceEvent:
    """An event emitted by a service."""

    def __init__(self, service: Service):
        self.service = service


class ServiceInitializedEvent(ServiceEvent):
    """This event is emitted every time a new service has been initialized."""

    def __init__(self, name: str, service: Service):
        super().__init__(service)
        self.name = name


class ServiceShutdownEvent(ServiceEvent):
    """This event is emitted every time a service has been shut down."""

    def __init__(self, name: str, service: Service):
        super().__init__(service)
        self.name = name


class ApplicationShutdown(ServiceEvent):
    """This event is emitted from the application when it has been shut
    down."""


class SessionCreated(ServiceEvent):
    """The session is emitting this event when it's ready to shut down."""

    def __init__(
        self,
        application: Service,
        session: Service,
        filename: str | Path | None,
        template: str | None,
    ):
        super().__init__(application)
        self.application = application
        self.session = session
        self.filename = Path(filename) if filename else None
        self.template = template


class ActiveSessionChanged(ServiceEvent):
    """Event emitted when a session becomes the active session."""


class SessionShutdownRequested(ServiceEvent):
    """When the application is asked to terminate, it will inform all sessions.

    The user can then save his/her work.
    """


class SessionShutdown(ServiceEvent):
    """The session is emitting this event when it's ready to shut down."""


class ModelLoaded:
    def __init__(self, service, filename: Path | None = None):
        self.service = service
        self.filename = filename


class ModelSaved:
    def __init__(self, service, filename: Path | None = None):
        self.service = service
        self.filename = filename


class TransactionBegin:
    """This event denotes the beginning of a transaction.

    Nested (sub-) transactions should not emit this signal.
    """


class TransactionCommit:
    """This event is emitted when a transaction (toplevel) is successfully
    committed."""


class TransactionRollback:
    """This event is emitted to tell the operation has failed.

    If a set of operations fail (e.i. due to an exception) the
    transaction should be marked for rollback.
    """


class ActionEnabled:
    """Signal if an action can be activated or not."""

    def __init__(self, action_name: str, enabled: bool) -> None:
        self.scope, self.name = (
            action_name.split(".", 2) if "." in action_name else ("win", action_name)
        )
        self.enabled = enabled


class Notification:
    """Inform the user about important events."""

    def __init__(self, message):
        self.message = message



""""""

from __future__ import annotations

import logging

from gaphor.abc import Service
from gaphor.core import event_handler
from gaphor.core.modeling.element import Element, Handler
from gaphor.core.modeling.event import (
    AssociationAdded,
    AssociationDeleted,
    AssociationSet,
    ElementUpdated,
    ModelReady,
)
from gaphor.core.modeling.properties import umlproperty

log = logging.getLogger(__name__)


class EventWatcher:
    """A helper for easy registering and unregistering event handlers."""

    def __init__(
        self,
        element: Element,
        element_dispatcher: ElementDispatcher | None,
        default_handler: Handler | None = None,
    ):
        self.element = element
        self.element_dispatcher = element_dispatcher
        self.default_handler: Handler | None = default_handler
        self._watched_paths: dict[str, Handler] = {}

    def watch(self, path: str, handler: Handler | None = None) -> EventWatcher:
        """Watch a certain path of elements starting with the DiagramItem. The
        handler is optional and will default the default provided at
        construction time.

        Watches should be set in the constructor, so they can be registered
        and unregistered in one shot.

        This interface is fluent (returns self).
        """
        if handler:
            self._watched_paths[path] = handler
        elif self.default_handler:
            self._watched_paths[path] = self.default_handler
        else:
            raise ValueError(f"No handler provided for path {path}")

        if dispatcher := self.element_dispatcher:
            dispatcher.subscribe(self._watched_paths[path], self.element, path)
        return self

    def unsubscribe_all(self, *_args):
        """Unregister handlers.

        Extra arguments are ignored (makes connecting to destroy signals
        much easier though).
        """
        dispatcher = self.element_dispatcher
        if not dispatcher:
            return

        for path, handler in self._watched_paths.items():
            dispatcher.unsubscribe(handler)


class ElementDispatcher(Service):
    """The Element based Dispatcher allows handlers to receive only events
    related to certain elements. Those elements should be registered too. A
    path should be provided, that is used to find those changes.

    The handlers are registered on their property attribute. This avoids
    subclass lookups and is pretty specific. As a result this dispatcher is
    tailored for dispatching events from the data model (ElementUpdated)

    For example: if you're a TransitionItem (gaphor.core.modeling.Presentation
    instance) and you're interested in the value of the guard attribute of the
    model element that's represented by this item (gaphor.UML.Transition), you
    can register a handler like this::

      dispatcher.subscribe(element,
              'guard.specification[LiteralSpecification].value', self._handler)

    Note the '[' and ']'. This is because guard references ValueSpecification,
    which does not have a value attribute. Therefore, the default reference type
    is overruled in favour of the LiteralSpecification.

    This dispatcher keeps track of the kind of events that are dispatched. The
    dispatcher table is updated accordingly (so the right handlers are fired
    every time).
    """

    def __init__(self, event_manager, modeling_language):
        self.event_manager = event_manager
        self.modeling_language = modeling_language

        # Table used to fire events:
        # (event.element, event.property): { handler: set(path, ..), ..}
        self._handlers: dict[tuple[Element, umlproperty], dict[Handler, set]] = {}

        # Fast resolution when handlers are disconnected
        # handler: [(element, property), ..]
        self._reverse: dict[Handler, list[tuple[Element, umlproperty]]] = {}

        self.event_manager.subscribe(self.on_model_loaded)
        self.event_manager.subscribe(self.on_element_change_event)

    def shutdown(self) -> None:
        self.event_manager.unsubscribe(self.on_element_change_event)
        self.event_manager.unsubscribe(self.on_model_loaded)

    def subscribe(self, handler: Handler, element: Element, path: str) -> None:
        props = self._path_to_properties(element, path)
        self._add_handlers(element, props, handler)

    def unsubscribe(self, handler: Handler) -> None:
        """Unregister a handler from the registry."""
        try:
            reverse = reversed(self._reverse[handler])
        except KeyError:
            return

        for key in reverse:
            if handlers := self._handlers.get(key):
                handlers.pop(handler, None)
                if not handlers:
                    del self._handlers[key]
        del self._reverse[handler]

    def _path_to_properties(self, element, path):
        """Given a start element and a path, return a tuple of properties
        (association, attribute, etc.) representing the path."""
        c = type(element)
        tpath = []
        for attr in path.split("."):
            cname = ""
            if "[" in attr:
                assert attr.endswith("]"), f'"{attr}" should end with "]"'
                attr, cname = attr[:-1].split("[")
            prop = getattr(c, attr)
            tpath.append(prop)

            if cname:
                c = self.modeling_language.lookup_element(cname)
                assert c and issubclass(
                    c, prop.type
                ), f"{c} should be a subclass of {prop.type}"
            else:
                c = prop.type
        return tuple(tpath)

    def _add_handlers(self, element, props, handler):
        """Provided an element and a path of properties (props), register the
        handler for each property."""
        property, remainder = props[0], props[1:]
        key = (element, property)

        # Register key
        try:
            handlers = self._handlers[key]
        except KeyError:
            handlers = {}
            self._handlers[key] = handlers

        # Register handler and it's remaining paths
        try:
            remainders = handlers[handler]
        except KeyError:
            remainders = handlers[handler] = set()
        if remainder:
            remainders.add(remainder)

        # Also add them to the reverse table, easing disconnecting
        try:
            reverse = self._reverse[handler]
        except KeyError:
            reverse = []
            self._reverse[handler] = reverse

        reverse.append(key)

        # Apply remaining path
        if remainder:
            if property.upper == "*" or property.upper > 1:
                for e in property.get(element):
                    self._add_handlers(e, remainder, handler)
            else:
                e = property.get(element)
                if e and remainder:
                    self._add_handlers(e, remainder, handler)

    def _remove_handlers(self, element, property, handler):
        """Remove the handler of the path of elements."""
        key = element, property
        handlers = self._handlers.get(key)
        if not handlers:
            return

        if property.upper == "*" or property.upper > 1:
            for remainder in handlers.get(handler, ()):
                for e in property.get(element):
                    self._remove_handlers(e, remainder[0], handler)
        else:
            for remainder in handlers.get(handler, ()):
                if e := property.get(element):
                    self._remove_handlers(e, remainder[0], handler)
        try:
            del handlers[handler]
        except KeyError:
            log.debug(
                "Handler %s is not registered for %s.%s",
                handler,
                element,
                property,
                exc_info=True,
            )

        if not handlers:
            del self._handlers[key]

    @event_handler(ElementUpdated)
    def on_element_change_event(self, event):
        if not (handlers := self._handlers.get((event.element, event.property))):
            return
        try:
            for handler in set(handlers.keys()):
                handler(event)
        finally:
            # Handle add/removal of handlers based on the kind of event
            # Filter out handlers that have no remaining properties
            if (
                isinstance(event, (AssociationSet, AssociationDeleted))
                and event.old_value
            ):
                for handler, remainders in handlers.items():
                    for remainder in remainders:
                        self._remove_handlers(event.old_value, remainder[0], handler)

            if (
                isinstance(event, (AssociationSet, AssociationAdded))
                and event.new_value
            ):
                for handler, remainders in handlers.items():
                    for remainder in remainders:
                        self._add_handlers(event.new_value, remainder, handler)

    @event_handler(ModelReady)
    def on_model_loaded(self, event):
        for (elem, prop), value in list(self._handlers.items()):
            prefix = (prop,)
            for h, remainders in list(value.items()):
                for remainder in remainders:
                    self._add_handlers(elem, prefix + remainder, h)



"""State items property pages.

To register property pages implemented in this module, it is imported in
gaphor.adapter package.
"""

from __future__ import annotations

from gaphor import UML
from gaphor.core import transactional
from gaphor.diagram.propertypages import (
    PropertyPageBase,
    PropertyPages,
    handler_blocking,
    new_resource_builder,
)
from gaphor.UML.states.state import StateItem
from gaphor.UML.states.statemachine import StateMachineItem

new_builder = new_resource_builder("gaphor.UML.states")


@PropertyPages.register(UML.Transition)
class TransitionPropertyPage(PropertyPageBase):
    """Transition property page allows to edit guard specification."""

    order = 15

    subject: UML.Transition

    def __init__(self, subject):
        self.subject = subject
        self.watcher = subject.watcher()

    def construct(self):
        subject = self.subject
        if not subject:
            return

        builder = new_builder(
            "transition-editor",
            signals={
                "transition-destroy": (self.watcher.unsubscribe_all,),
            },
        )

        guard = builder.get_object("guard")
        if subject.guard:
            guard.set_text(subject.guard.specification or "")

        @handler_blocking(guard, "changed", self._on_guard_change)
        def handler(event):
            if event.element is subject.guard and guard.get_text() != event.new_value:
                guard.set_text(event.new_value or "")

        self.watcher.watch("guard[Constraint].specification", handler)

        return builder.get_object("transition-editor")

    @transactional
    def _on_guard_change(self, entry):
        value = entry.get_text().strip()
        if not self.subject.guard:
            self.subject.guard = self.subject.model.create(UML.Constraint)
        self.subject.guard.specification = value


@PropertyPages.register(UML.State)
class StatePropertyPage(PropertyPageBase):
    """State property page."""

    order = 15
    subject: UML.State

    def __init__(self, subject):
        self.subject = subject

    def construct(self):
        subject = self.subject
        if not subject:
            return

        builder = new_builder(
            "state-editor",
            signals={
                "entry-changed": (self.on_text_change, self.set_entry),
                "exit-changed": (self.on_text_change, self.set_exit),
                "do-activity-changed": (self.on_text_change, self.set_do_activity),
            },
        )

        entry = builder.get_object("entry")
        if subject.entry:
            entry.set_text(subject.entry.name or "")

        exit = builder.get_object("exit")
        if subject.exit:
            exit.set_text(subject.exit.name or "")

        do_activity = builder.get_object("do-activity")
        if subject.doActivity:
            do_activity.set_text(self.subject.doActivity.name or "")

        return builder.get_object("state-editor")

    @transactional
    def on_text_change(self, entry, method):
        value = entry.get_text().strip()
        method(value)

    def set_entry(self, text):
        if not self.subject.entry:
            self.subject.entry = self.subject.model.create(UML.Activity)
        self.subject.entry.name = text

    def set_exit(self, text):
        if not self.subject.exit:
            self.subject.exit = self.subject.model.create(UML.Activity)
        self.subject.exit.name = text

    def set_do_activity(self, text):
        if not self.subject.doActivity:
            self.subject.doActivity = self.subject.model.create(UML.Activity)
        self.subject.doActivity.name = text


@PropertyPages.register(StateItem)
@PropertyPages.register(StateMachineItem)
class RegionPropertyPage(PropertyPageBase):

    order = 15

    def __init__(self, item: StateItem | StateMachineItem):
        super().__init__()
        self.item = item

    def construct(self):
        builder = new_builder(
            "region-editor",
            "num-regions-adjustment",
            signals={
                "show-regions-changed": (self._on_show_regions_changed,),
                "regions-changed": (self._on_num_regions_changed,),
            },
        )

        num_regions = builder.get_object("num-regions")
        num_regions.set_value(len(self.item.subject.region))

        show_regions = builder.get_object("show-regions")
        show_regions.set_active(self.item.show_regions)

        return builder.get_object("region-editor")

    @transactional
    def _on_num_regions_changed(self, spin_button):
        num_regions = spin_button.get_value_as_int()
        self.update_regions(num_regions)

    @transactional
    def _on_show_regions_changed(self, button, gparam):
        self.item.show_regions = button.get_active()

    def update_regions(self, num_regions) -> None:
        if not self.item.subject:
            return

        while num_regions > len(self.item.subject.region):
            region = self.item.subject.model.create(UML.Region)
            self.item.subject.region = region

        while num_regions < len(self.item.subject.region):
            region = self.item.subject.region[-1]
            region.unlink()



"""Implementation of connector from Composite Structures and Components.

Only assembly connector (see chapter Components in UML specification) is
supported at the moment. The implementation is based on `ConnectorItem`
class and `InterfaceItem` class in assembly connector mode.

Assembly Connector
==================
To connect two components with assembly connector connect folded interface
and component items using connector item.

If component provides or requires a connected interface, then assembly
connection in UML data model will be created and connector item will
display name of the interface. Otherwise, UML data model is not updated
and connector item does not display interface name.

Interface item in assembly connector mode does not display interface name
as it is displayed by connectors.

Connector item visualizes two UML metaclasses

- ConnectorEnd metaclass when connecting to interface item in assembly mode
- Connector metaclass in other cases

Using property pages of connector item, user can change superinterface of
connected interface.

Assembly Connector Mode of Interface Item
-----------------------------------------
Assembly connector notation is supported using interface item because of
its simplicity

- no need for additional assembly connector item
- because connection is made to specific interface, there is no need for
  performing a search for common interface of all connected components
- separate assembly connector item would require some rotation support,
  instead interface item's rotation capabilities are reused

Implementation Alternatives
---------------------------
There were several alternatives of assembly connector notation explored.

In Gaphor 0.8.x there was assembly connector item, with additional handles
and lines. User was dragging a handle of an additional line to connect to
a component, disadvantages

- item's connection behaviour is not consistent with other items
- rotation needs to be implemented

For Gaphor 0.14 and later, two other ideas were considered.

First one required assembly connector item as well. Connector item could
visualize ConnectorEnd and Connector UML metaclasses, and it would be
used to connect assembly connector item and items of components. It is very
consistent with the rest of Gaphor application but

- it proved to be very complicated in implementation
- requires additional item

Second alternative was to have connector item only. It is very simple
concept in first place. When connector item connects two components, then
draw assembly connector icon in the middle of a line. The solution is very
simple in implementation and consistent with the rest of the application
until multiple components have to be connected with one assembly connector.

UML Specification Issues
========================
UML specification is not clear about interfaces as connectable elements
and connector's `kind` attribute.

Current implementation is subject to change in the future, when UML
specification clarifies issues described below.

See also http://www.omg.org/issues/uml2-rtf.open.html#Issue7251

Connector Kind
--------------
Chapter Components of UML specification adds `kind` attribute to connector
metaclass. This is enumeration with two possible values `assembly' and
`delegation'.

It is not clear what value should be assigned to `kind` attribute of
connector, which is defined between connectable elements like ports (not
characterized by interfaces), properties and parameters.

Interfaces as Connectable Elements
----------------------------------
Chapter Composite Structures in UML Superstructure 2.1.2 document does not
specify interfaces as connectable elements.

But definition of assembly connector says:

    An assembly connector is a connector between two components that
    defines that one component provides the services that another component
    requires. An assembly connector is a connector that is defined from
    a required _interface_ or port to a provided _interface_ or port.

Therefore, code of connector items is written with assumption, that
interfaces are connectable elements.
"""

from gaphor import UML
from gaphor.core.format import format
from gaphor.diagram.presentation import LinePresentation, Named
from gaphor.diagram.shapes import Box, Text, cairo_state
from gaphor.diagram.support import represents
from gaphor.UML.classes.association import get_center_pos
from gaphor.UML.recipes import stereotypes_str


@represents(UML.Connector)
class ConnectorItem(Named, LinePresentation[UML.Connector]):
    """Connector item line.

    Represents Connector UML metaclass. If connected to interface item in
    assembly connector mode, then `Connector.end` attribute represents
    appropriate `ConnectorEnd` UML metaclass instance.

    :Attributes:
     subject
        Connector UML metaclass instance.
     end
        ConnectorEnd UML metaclass instance.
    """

    def __init__(self, diagram, id=None):
        super().__init__(
            diagram,
            id,
            shape_middle=Box(
                Text(
                    text=lambda: stereotypes_str(self.subject),
                ),
                Text(text=lambda: self.subject.name or ""),
                Text(
                    text=lambda: ", ".join(
                        self.subject.informationFlow[:].conveyed[:].name
                    )
                ),
                # Also support SysML ItemFlow:
                Text(
                    text=lambda: stereotypes_str(
                        self.subject.informationFlow[0].itemProperty.type,  # type: ignore[attr-defined]
                        raaml_stereotype_workaround(
                            self.subject.informationFlow[0].itemProperty.type  # type: ignore[attr-defined]
                        ),
                    )
                    if self.subject.informationFlow
                    else ""
                ),
                Text(
                    text=lambda: format(
                        self.subject.informationFlow[0].itemProperty, type=True  # type: ignore[attr-defined]
                    )
                    if self.subject.informationFlow
                    else ""
                ),
            ),
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject.appliedStereotype.classifier.name")
        self.watch("subject[Connector].informationFlow.informationSource")
        self.watch("subject[Connector].informationFlow.conveyed.name")
        self.watch("subject[Connector].informationFlow[ItemFlow].itemProperty.name")
        self.watch(
            "subject[Connector].informationFlow[ItemFlow].itemProperty.type.name"
        )
        self.watch(
            "subject[Connector].informationFlow[ItemFlow].itemProperty.type.appliedStereotype.classifier.name"
        )

    def draw(self, context):
        super().draw(context)
        subject = self.subject
        if subject and subject.informationFlow:

            inv = (
                1
                if (subject.end[0].role in subject.informationFlow[:].informationTarget)
                else -1
            )
            handles = self.handles()
            pos, angle = get_center_pos(handles)
            with cairo_state(context.cairo) as cr:
                cr.translate(*pos)
                cr.rotate(angle)
                cr.move_to(0, 0)
                cr.line_to(12 * inv, 8)
                cr.line_to(12 * inv, -8)
                cr.fill()

    def draw_tail(self, context):
        cr = context.cairo
        cr.line_to(0, 0)
        if self.subject and self.subject.kind == "delegation":
            cr.move_to(15, -6)
            cr.line_to(0, 0)
            cr.line_to(15, 6)


def raaml_stereotype_workaround(element):
    """This is a temporary fix to ensure ItemFlow is showing proper stereotypes
    for Controller, Feedback and ControlAction from RAAML."""
    if not element:
        return ()

    name: str = type(element).__name__
    if name in {"Controller", "Feedback", "ControlAction"}:
        return (name[0].lower() + name[1:],)
    return ()



"""This module has a generic file dialog functions that are used to open or
save files."""

from __future__ import annotations

import sys
from pathlib import Path

from gi.repository import Gio, Gtk

from gaphor.i18n import gettext

GAPHOR_FILTER = [(gettext("All Gaphor Models"), "*.gaphor", "application/x-gaphor")]


def new_filter(name, pattern, mime_type=None):
    f = Gtk.FileFilter.new()
    f.set_name(name)
    f.add_pattern(pattern)
    if mime_type:
        f.add_mime_type(mime_type)
    return f


def _file_dialog_with_filters(title, parent, action, filters):
    dialog = Gtk.FileChooserNative.new(title, parent, action, None, None)

    if parent:
        dialog.set_transient_for(parent)

    if sys.platform != "darwin":
        for name, pattern, mime_type in filters:
            dialog.add_filter(new_filter(name, pattern, mime_type))
        dialog.add_filter(new_filter(gettext("All Files"), "*"))
    return dialog


def open_file_dialog(title, handler, parent=None, dirname=None, filters=None) -> None:
    if filters is None:
        filters = []
    dialog = _file_dialog_with_filters(
        title, parent, Gtk.FileChooserAction.OPEN, filters
    )
    dialog.set_select_multiple(True)

    def response(_dialog, answer):
        if Gtk.get_major_version() == 3:
            filenames = (
                dialog.get_filenames() if answer == Gtk.ResponseType.ACCEPT else []
            )
        else:
            filenames = (
                [f.get_path() for f in dialog.get_files()]
                if answer == Gtk.ResponseType.ACCEPT
                else []
            )
        dialog.destroy()
        handler(filenames)

    dialog.connect("response", response)
    dialog.set_modal(True)
    if Gtk.get_major_version() == 3:
        if dirname:
            dialog.set_current_folder(dirname)
    else:
        if dirname:
            dialog.set_current_folder(Gio.File.parse_name(dirname))
    dialog.show()


def save_file_dialog(
    title,
    handler,
    parent=None,
    filename=None,
    extension=None,
    filters=None,
) -> None:
    if filters is None:
        filters = []
    dialog = _file_dialog_with_filters(
        title, parent, Gtk.FileChooserAction.SAVE, filters
    )

    def get_filename() -> Path:
        if Gtk.get_major_version() == 3:
            return Path(dialog.get_filename())
        else:
            return Path(dialog.get_file().get_path())

    def set_filename(filename: Path):
        if Gtk.get_major_version() == 3:
            dialog.set_filename(str(filename.name))
        else:
            dialog.set_current_name(str(filename.name))

    def overwrite_check() -> Path | None:
        filename = get_filename()
        if extension and filename.suffix != extension:
            filename = filename.with_suffix(extension)
            set_filename(filename)
            return None if filename.exists() else filename
        return filename

    def response(_dialog, answer):
        if answer == Gtk.ResponseType.ACCEPT:
            if filename := overwrite_check():
                dialog.destroy()
                handler(filename)
            else:
                dialog.show()
        else:
            dialog.destroy()

    dialog.connect("response", response)
    if filename:
        set_filename(filename)
    if Gtk.get_major_version() == 3:
        dialog.set_do_overwrite_confirmation(True)
        if filename:
            dialog.set_current_folder(str(filename.parent))
    else:
        if filename:
            dialog.set_current_folder(Gio.File.parse_name(str(filename.parent)))
    dialog.set_modal(True)
    dialog.show()



"""
Association item - graphical representation of an association.

Plan:
 - transform AssociationEnd in a (dumb) data class
 - for association name and direction tag, use the same trick as is used
   for line ends.
"""

from math import atan2, pi
from typing import Optional

from gaphas.connector import Handle
from gaphas.geometry import Rectangle, distance_rectangle_point

from gaphor import UML
from gaphor.core.modeling.properties import association, attribute, enumeration
from gaphor.core.styling import Style, merge_styles
from gaphor.diagram.presentation import LinePresentation, Named
from gaphor.diagram.shapes import (
    Box,
    Text,
    cairo_state,
    draw_default_head,
    draw_default_tail,
    stroke,
)
from gaphor.diagram.support import represents
from gaphor.diagram.text import Layout, middle_segment
from gaphor.UML.recipes import stereotypes_str
from gaphor.UML.umlfmt import format_association_end

half_pi = pi / 2


@represents(UML.Association)
class AssociationItem(Named, LinePresentation[UML.Association]):
    """AssociationItem represents associations.

    An AssociationItem has two AssociationEnd items. Each AssociationEnd
    item represents a Property (with Property.association == my
    association).
    """

    def __init__(self, diagram, id=None):
        super().__init__(
            diagram,
            id,
            shape_middle=Box(
                Text(
                    text=lambda: stereotypes_str(self.subject),
                ),
                Text(text=lambda: self.subject.name or ""),
            ),
        )

        # AssociationEnds are really inseparable from the AssociationItem.
        # We give them the same id as the association item.
        self._head_end = AssociationEnd(owner=self, end="head")
        self._tail_end = AssociationEnd(owner=self, end="tail")

        # For the association ends:
        base = "subject[Association].memberEnd[Property]"
        self.watch("subject[NamedElement].name").watch(
            "subject.appliedStereotype.classifier.name"
        ).watch(f"{base}.name", self.on_association_end_value).watch(
            f"{base}.aggregation", self.on_association_end_value
        ).watch(
            f"{base}.appliedStereotype.slot.definingFeature.name",
            self.on_association_end_value,
        ).watch(
            f"{base}.appliedStereotype.slot.value", self.on_association_end_value
        ).watch(
            f"{base}.classifier", self.on_association_end_value
        ).watch(
            f"{base}.visibility", self.on_association_end_value
        ).watch(
            f"{base}.lowerValue", self.on_association_end_value
        ).watch(
            f"{base}.upperValue", self.on_association_end_value
        ).watch(
            f"{base}.owningAssociation", self.on_association_end_value
        ).watch(
            f"{base}.type[Class].ownedAttribute", self.on_association_end_value
        ).watch(
            f"{base}.type[Interface].ownedAttribute", self.on_association_end_value
        ).watch(
            f"{base}.appliedStereotype.classifier", self.on_association_end_value
        ).watch(
            "subject[Association].memberEnd"
        ).watch(
            "subject[Association].ownedEnd"
        ).watch(
            "subject[Association].navigableOwnedEnd"
        ).watch(
            "show_direction"
        ).watch(
            "preferred_aggregation", self.on_association_end_value
        )

    show_direction: attribute[int] = attribute("show_direction", int, default=False)
    preferred_aggregation = enumeration(
        "preferred_aggregation", ("none", "shared", "composite"), "none"
    )

    def load(self, name, value):
        # end_head and end_tail were used in an older Gaphor version
        if name in ("head_end", "head-subject"):
            name = "head_subject"
        elif name in ("tail_end", "tail-subject"):
            name = "tail_subject"
        elif name == "show-direction":
            name = "show_direction"

        super().load(name, value)

    def postload(self):
        super().postload()
        self.on_association_end_value()

    head_end = property(lambda self: self._head_end)
    tail_end = property(lambda self: self._tail_end)
    head_subject = association("head_subject", UML.Property, upper=1)
    tail_subject = association("tail_subject", UML.Property, upper=1)

    def invert_direction(self):
        """Invert the direction of the association, this is done by swapping
        the head and tail-ends subjects."""
        if not self.subject:
            return

        self.subject.memberEnd.swap(
            self.subject.memberEnd[0], self.subject.memberEnd[1]
        )
        self.request_update()

    def update_ends(self):
        self.on_association_end_value()

    def on_association_end_value(self, event=None):
        """Handle events and update text on association end."""
        for end in (self._head_end, self._tail_end):
            end.set_text()

        # Update line endings:
        head_subject = self.head_subject
        tail_subject = self.tail_subject

        # Update line ends using the aggregation and isNavigable values:
        if head_subject and tail_subject:
            if tail_subject.aggregation == "composite":
                self.draw_head = draw_head_composite
            elif tail_subject.aggregation == "shared":
                self.draw_head = draw_head_shared
            elif head_subject.navigability is True:
                self.draw_head = draw_head_navigable
            elif head_subject.navigability is False:
                self.draw_head = draw_head_none
            else:
                self.draw_head = draw_default_head

            if head_subject.aggregation == "composite":
                self.draw_tail = draw_tail_composite
            elif head_subject.aggregation == "shared":
                self.draw_tail = draw_tail_shared
            elif tail_subject.navigability is True:
                self.draw_tail = draw_tail_navigable
            elif tail_subject.navigability is False:
                self.draw_tail = draw_tail_none
            else:
                self.draw_tail = draw_default_tail
        else:
            if self.preferred_aggregation == "composite":
                self.draw_head = draw_head_composite
            elif self.preferred_aggregation == "shared":
                self.draw_head = draw_head_shared
            else:
                self.draw_head = draw_default_head
            self.draw_tail = draw_default_tail

        self.request_update()

    def update(self, context):
        self.update_ends()

    def point(self, x, y):
        """Returns the distance from the Association to the (mouse) cursor."""
        return min(
            super().point(x, y), self._head_end.point(x, y), self._tail_end.point(x, y)
        )

    def draw(self, context):
        super().draw(context)

        handles = self.handles()

        # Calculate alignment of the head name and multiplicity
        self._head_end.update_position(context, handles[0].pos, handles[1].pos)

        # Calculate alignment of the tail name and multiplicity
        self._tail_end.update_position(context, handles[-1].pos, handles[-2].pos)

        self._head_end.draw(context)
        self._tail_end.draw(context)
        if self.show_direction:
            pos, angle = get_center_pos(handles)
            inv = (
                -1
                if (
                    self.tail_subject and self.tail_subject is self.subject.memberEnd[0]
                )
                else 1
            )
            with cairo_state(context.cairo) as cr:
                cr.translate(*pos)
                if -half_pi <= angle < half_pi:
                    angle += pi
                    inv *= -1
                cr.rotate(angle)
                cr.move_to(0, 2)
                cr.line_to(6 * inv, 7)
                cr.line_to(0, 12)
                cr.fill()


def get_center_pos(points):
    """Return position in the centre of middle segment of a line.

    Angle of the middle segment is also returned.
    """
    h0, h1 = middle_segment(points)
    pos = (h0.pos.x + h1.pos.x) / 2, (h0.pos.y + h1.pos.y) / 2
    angle = atan2(h1.pos.y - h0.pos.y, h1.pos.x - h0.pos.x)
    return pos, angle


def draw_head_none(context):
    """Draw an 'x' on the line end to indicate no navigability at association
    head."""
    cr = context.cairo
    cr.move_to(6, -4)
    cr.rel_line_to(8, 8)
    cr.rel_move_to(0, -8)
    cr.rel_line_to(-8, 8)
    cr.move_to(0, 0)


def draw_tail_none(context):
    """Draw an 'x' on the line end to indicate no navigability at association
    tail."""
    cr = context.cairo
    cr.line_to(0, 0)
    cr.move_to(6, -4)
    cr.rel_line_to(8, 8)
    cr.rel_move_to(0, -8)
    cr.rel_line_to(-8, 8)


def _draw_diamond(cr):
    """Helper function to draw diamond shape for shared and composite
    aggregations."""
    cr.move_to(20, 0)
    cr.line_to(10, -6)
    cr.line_to(0, 0)
    cr.line_to(10, 6)
    cr.close_path()


def draw_head_composite(context):
    """Draw a closed diamond on the line end to indicate composite aggregation
    at association head."""
    cr = context.cairo
    _draw_diamond(cr)
    context.cairo.fill_preserve()
    cr.move_to(20, 0)


def draw_tail_composite(context):
    """Draw a closed diamond on the line end to indicate composite aggregation
    at association tail."""
    cr = context.cairo
    cr.line_to(20, 0)
    stroke(context)
    _draw_diamond(cr)
    cr.fill_preserve()
    stroke(context)


def draw_head_shared(context):
    """Draw an open diamond on the line end to indicate shared aggregation at
    association head."""
    cr = context.cairo
    _draw_diamond(cr)
    cr.move_to(20, 0)


def draw_tail_shared(context):
    """Draw an open diamond on the line end to indicate shared aggregation at
    association tail."""
    cr = context.cairo
    cr.line_to(20, 0)
    _draw_diamond(cr)


def draw_head_navigable(context):
    """Draw a normal arrow to indicate association end navigability at
    association head."""
    cr = context.cairo
    cr.move_to(15, -6)
    cr.line_to(0, 0)
    cr.line_to(15, 6)
    cr.move_to(0, 0)


def draw_tail_navigable(context):
    """Draw a normal arrow to indicate association end navigability at
    association tail."""
    cr = context.cairo
    cr.line_to(0, 0)
    cr.move_to(15, -6)
    cr.line_to(0, 0)
    cr.line_to(15, 6)


class AssociationEnd:
    """An association end represents one end of an association. An association
    has two ends. An association end has two labels: one for the name and one
    for the multiplicity (and maybe one for tagged values in the future).

    An AsociationEnd has no ID, hence it will not be stored, but it will
    be recreated by the owning Association.
    """

    def __init__(self, owner: AssociationItem, end: Optional[str] = None):
        self._canvas = None
        self._owner = owner
        self._end = end

        # Rendered text for name and multiplicity
        self._name = ""
        self._mult = ""

        self._name_bounds = Rectangle()
        self._mult_bounds = Rectangle()

        self._name_layout = Layout("")
        self._mult_layout = Layout("")

        self._inline_style: Style = {"font-size": "x-small"}

    name_bounds = property(lambda s: s._name_bounds)

    @property
    def owner(self) -> AssociationItem:
        """Override Element.owner."""
        return self._owner

    @property
    def owner_handle(self) -> Handle:
        # handle(event) is the event handler method
        return self._owner.head if self is self._owner.head_end else self._owner.tail

    @property
    def subject(self) -> Optional[UML.Property]:
        return getattr(self.owner, f"{self._end}_subject")  # type:ignore[no-any-return]

    def request_update(self):
        self._owner.request_update()

    def set_text(self):
        """Set the text on the association end."""
        if self.subject:
            try:
                n, m = format_association_end(self.subject)
            except ValueError:
                # need more than 0 values to unpack: property was rendered as
                # attribute while in a UNDO action for example.
                pass
            else:
                self._name = n
                self._mult = m
                self.request_update()

    def get_name(self):
        return self._name

    def get_mult(self):
        return self._mult

    def update_position(self, context, p1, p2):
        """Update label placement for association's name and multiplicity
        label.

        p1 is the line end and p2 is the last but one point of the line.
        """
        style = merge_styles(context.style, self._inline_style)
        ofs = 4.0

        dx = float(p2[0]) - float(p1[0])
        dy = float(p2[1]) - float(p1[1])

        def max_text_size(size1, size2):
            w1, h1 = size1
            w2, h2 = size2
            return (max(w1, w2), max(h1, h2))

        name_layout = self._name_layout
        name_layout.set_text(self._name)
        name_layout.set_font(style)
        name_w, name_h = max_text_size(name_layout.size(), (10, 10))

        mult_layout = self._mult_layout
        mult_layout.set_text(self._mult)
        mult_layout.set_font(style)
        mult_w, mult_h = max_text_size(mult_layout.size(), (10, 10))

        rc = 1000.0 if dy == 0 else dx / dy
        abs_rc = abs(rc)
        h = dx > 0  # right side of the box
        v = dy > 0  # bottom side

        if abs_rc > 6:
            # horizontal line
            if h:
                name_dx = ofs
                name_dy = -ofs * 2 - name_h
                mult_dx = ofs
                mult_dy = ofs * 2
            else:
                name_dx = -ofs - name_w
                name_dy = -ofs - name_h
                mult_dx = -ofs - mult_w
                mult_dy = ofs
        elif 0 <= abs_rc <= 0.2:
            # vertical line
            if v:
                name_dx = -ofs * 2.5 - name_w
                name_dy = ofs
                mult_dx = ofs * 2.5
                mult_dy = ofs
            else:
                name_dx = -ofs - name_w
                name_dy = -ofs - name_h
                mult_dx = ofs
                mult_dy = -ofs - mult_h
        else:
            # Should both items be placed on the same side of the line?
            r = abs_rc < 1.0

            # Find out alignment of text (depends on the direction of the line)
            align_left = h ^ r
            align_bottom = v ^ r
            if align_left:
                name_dx = ofs
                mult_dx = ofs
            else:
                name_dx = -ofs - name_w
                mult_dx = -ofs - mult_w
            if align_bottom:
                name_dy = -ofs - name_h
                mult_dy = -ofs - name_h - mult_h
            else:
                name_dy = ofs
                mult_dy = ofs + mult_h

        self._name_bounds = Rectangle(
            p1[0] + name_dx, p1[1] + name_dy, width=name_w, height=name_h
        )

        self._mult_bounds = Rectangle(
            p1[0] + mult_dx, p1[1] + mult_dy, width=mult_w, height=mult_h
        )

    def point(self, x, y):
        """Given a point (x, y) return the distance to the diagram item."""
        drp = distance_rectangle_point
        pos = (x, y)
        d1 = drp(self._name_bounds, pos)
        d2 = drp(self._mult_bounds, pos)
        d3 = 1000.0
        return min(d1, d2, d3)

    def draw(self, context):
        """Draw name and multiplicity of the line end."""
        if not self.subject:
            return

        cr = context.cairo
        if text_color := context.style.get("text-color"):
            cr.set_source_rgba(*text_color)

        cr.move_to(self._name_bounds.x, self._name_bounds.y)
        self._name_layout.show_layout(cr)
        cr.move_to(self._mult_bounds.x, self._mult_bounds.y)
        self._mult_layout.show_layout(cr)



from __future__ import annotations

import logging

from generic.event import Event
from gi.repository import Gio, Gtk

from gaphor.abc import ActionProvider
from gaphor.core import action, event_handler
from gaphor.core.modeling import AttributeUpdated, Diagram, ModelFlushed, ModelReady
from gaphor.diagram.drop import drop
from gaphor.event import ActionEnabled
from gaphor.transaction import Transaction
from gaphor.ui.abc import UIComponent
from gaphor.ui.diagrampage import DiagramPage, GtkView
from gaphor.ui.event import (
    CurrentDiagramChanged,
    DiagramClosed,
    DiagramOpened,
    DiagramSelectionChanged,
    ElementOpened,
)

if Gtk.get_major_version() != 3:
    from gi.repository import Adw

log = logging.getLogger(__name__)

DIAGRAM_ICON = Gio.ThemedIcon.new("gaphor-diagram-symbolic")


class Diagrams(UIComponent, ActionProvider):
    def __init__(
        self, event_manager, element_factory, properties, modeling_language, toolbox
    ):
        self.event_manager = event_manager
        self.element_factory = element_factory
        self.properties = properties
        self.modeling_language = modeling_language
        self.toolbox = toolbox
        self._notebook: Gtk.Widget = None
        self._page_handler_ids: list[int] = []

    def open(self):
        """Open the diagrams component.

        Returns:
            The Gtk.Notebook (GTK3) or Gtk.Box with Adw.TabView and Adw.TabBar (GTK4).
        """

        if Gtk.get_major_version() == 3:
            self._notebook = Gtk.Notebook()
            self._notebook.props.scrollable = True
            self._notebook.show()

            self._notebook.connect("destroy", self._on_notebook_destroy)
            self._notebook.connect("switch-page", self._on_switch_page)
            self._page_handler_ids = [
                self._notebook.connect("page-added", self._on_page_changed),
                self._notebook.connect("page-removed", self._on_page_changed),
                self._notebook.connect("page-reordered", self._on_page_changed),
                self._notebook.connect("notify::page", self._on_current_page_changed),
            ]
        else:
            self._notebook = Adw.TabView()
            self._notebook.props.vexpand = True
            self._bar = Adw.TabBar()
            self._bar.set_view(self._notebook)
            self._box = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)
            self._box.append(self._bar)
            self._box.append(self._notebook)
            # connect view::close-page
            self._page_handler_ids = [
                self._notebook.connect(
                    "close-page",
                    lambda _notebook, page: self.event_manager.handle(
                        DiagramClosed(page.get_child().diagram_page.get_diagram())
                    ),
                ),
                self._notebook.connect("page-attached", self._on_page_changed),
                self._notebook.connect("page-detached", self._on_page_changed),
                self._notebook.connect("page-reordered", self._on_page_changed),
                self._notebook.connect(
                    "notify::selected-page", self._on_current_page_changed
                ),
            ]

        self.event_manager.subscribe(self._on_show_diagram)
        self.event_manager.subscribe(self._on_show_element)
        self.event_manager.subscribe(self._on_close_diagram)
        self.event_manager.subscribe(self._on_name_change)
        self.event_manager.subscribe(self._on_flush_model)
        self.event_manager.subscribe(self._on_model_ready)

        self._on_model_ready()
        return self._notebook if Gtk.get_major_version() == 3 else self._box

    def close(self):
        """Close the diagrams component."""

        self.event_manager.unsubscribe(self._on_model_ready)
        self.event_manager.unsubscribe(self._on_flush_model)
        self.event_manager.unsubscribe(self._on_name_change)
        self.event_manager.unsubscribe(self._on_close_diagram)
        self.event_manager.unsubscribe(self._on_show_element)
        self.event_manager.unsubscribe(self._on_show_diagram)
        if self._notebook:
            if Gtk.get_major_version() == 3:
                self._notebook.destroy()
            elif parent := self._notebook.get_parent():
                parent.remove(self._notebook)
            self._notebook = None

    def get_current_diagram(self) -> Diagram | None:
        """Returns the current page of the notebook.

        Returns (DiagramPage): The current diagram page.
        """
        if not self._notebook:
            return None
        if Gtk.get_major_version() == 3:
            page_num = self._notebook.get_current_page()
            child_widget = self._notebook.get_nth_page(page_num)
            return child_widget.diagram_page.get_diagram() if child_widget else None
        else:
            selected = self._notebook.get_selected_page()
            return selected.get_child().diagram_page.get_diagram() if selected else None

    def set_current_diagram(self, diagram: Diagram) -> bool:
        for page_num, widget in get_widgets_on_pages(self._notebook):
            if widget.diagram_page.get_diagram() is diagram:
                if Gtk.get_major_version() == 3:
                    self._notebook.set_current_page(page_num)
                else:
                    self._notebook.set_selected_page(
                        self._notebook.get_nth_page(page_num)
                    )
                self.get_current_view().grab_focus()
                return True
        return False

    def get_current_view(self) -> GtkView:
        """Returns the current view of the diagram page.

        Returns (GtkView): The current view.
        """
        if not self._notebook:
            return None
        if Gtk.get_major_version() == 3:
            page_num = self._notebook.get_current_page()
            child_widget = self._notebook.get_nth_page(page_num)
            return child_widget and child_widget.diagram_page.get_view()
        else:
            selected = self._notebook.get_selected_page()
            return selected and selected.get_child().diagram_page.get_view()

    def create_diagram_page(self, diagram: Diagram) -> DiagramPage:
        page = DiagramPage(
            diagram,
            self.event_manager,
            self.element_factory,
            self.properties,
            self.modeling_language,
        )
        widget = page.construct()
        widget.diagram_page = page

        apply_tool_select_controller(widget, self.toolbox)
        self._create_tab(diagram.name, widget)
        page.select_tool(self.toolbox.active_tool_name)
        self.get_current_view().grab_focus()
        self._update_action_state()
        return page

    def _create_tab(self, title, widget):
        """Creates a new Notebook tab with a label and close button.

        Args:
            title (str): The title of the tab, the diagram name.
            widget (Gtk.Widget): The child widget of the tab.
        """
        if Gtk.get_major_version() == 3:
            page_num = self._notebook.append_page(
                child=widget, tab_label=tab_label(title, widget, self.event_manager)
            )
            self._notebook.set_current_page(page_num)
            self._notebook.set_tab_reorderable(widget, True)
        else:
            page = self._notebook.append(widget)
            page.set_title(title or "")
            page.set_icon(DIAGRAM_ICON)
            self._notebook.set_selected_page(page)

        view = widget.diagram_page.view
        self.event_manager.handle(
            DiagramSelectionChanged(
                view, view.selection.focused_item, view.selection.selected_items
            )
        )

    def _on_notebook_destroy(self, notebook):
        for id in self._page_handler_ids:
            notebook.disconnect(id)

    def _on_switch_page(self, notebook, page, new_page_num):
        view = page.diagram_page.view
        self.event_manager.handle(
            DiagramSelectionChanged(
                view, view.selection.focused_item, view.selection.selected_items
            )
        )

    def _on_page_changed(self, notebook, _page, _page_num):
        def diagram_ids():
            notebook = self._notebook
            if not notebook:
                return
            for page_num in range(notebook.get_n_pages()):
                if page := notebook.get_nth_page(page_num):
                    if Gtk.get_major_version() != 3:
                        page = page.get_child()
                    if diagram := page.diagram_page.get_diagram():
                        yield diagram.id

        self.properties.set("opened-diagrams", list(diagram_ids()))

    def _on_current_page_changed(self, _notebook_or_tab_page, _gparam):
        diagram = self.get_current_diagram()
        self.event_manager.handle(CurrentDiagramChanged(diagram))
        self.properties.set("current-diagram", diagram.id if diagram else None)

    @action(name="close-current-tab", shortcut="<Primary>w")
    def close_current_tab(self):
        diagram = self.get_current_diagram()
        self.event_manager.handle(DiagramClosed(diagram))

    @action(
        name="zoom-in",
        shortcut=("<Primary>equal", "<Primary>plus"),
    )
    def zoom_in(self):
        if view := self.get_current_view():
            view.zoom(1.2)

    @action(
        name="zoom-out",
        shortcut="<Primary>minus",
    )
    def zoom_out(self):
        if view := self.get_current_view():
            view.zoom(1 / 1.2)

    @action(
        name="zoom-100",
        shortcut="<Primary>0",
    )
    def zoom_100(self):
        if view := self.get_current_view():
            zx = view.matrix[0]
            view.zoom(1 / zx)

    @action(
        name="select-all",
        shortcut="<Primary>a",
    )
    def select_all(self):
        view = self.get_current_view()
        if view and view.has_focus():
            view.selection.select_items(*view.model.get_all_items())

    @action(name="unselect-all", shortcut="<Primary><Shift>a")
    def unselect_all(self):
        view = self.get_current_view()
        if view and view.has_focus():
            view.selection.unselect_all()

    @event_handler(ElementOpened)
    def _on_show_element(self, event: ElementOpened):
        view = self.get_current_view()
        diagram = self.get_current_diagram()
        if not (view and diagram):
            return
        x, y = view.matrix.inverse().transform_point(100, 100)
        with Transaction(self.event_manager):
            if item := drop(event.element, diagram, x, y):
                view.selection.unselect_all()
                view.selection.focused_item = item

    @event_handler(DiagramOpened)
    def _on_show_diagram(self, event: DiagramOpened):
        """Show a Diagram element in the Notebook.

        If a diagram is already open on a Notebook page, show that one,
        otherwise create a new Notebook page.

        Args:
            event: The service event that is calling the method.
        """
        diagram = event.diagram
        if not self.set_current_diagram(diagram):
            self.create_diagram_page(diagram)

    @event_handler(DiagramClosed)
    def _on_close_diagram(self, event: Event) -> None:
        """Callback to close the tab and remove the notebook page."""

        diagram = event.diagram

        for page_num, widget in get_widgets_on_pages(self._notebook):
            if widget.diagram_page.get_diagram() is diagram:
                break
        else:
            return

        if Gtk.get_major_version() == 3:
            self._notebook.remove_page(page_num)
            widget.destroy()
        else:
            self._notebook.close_page(self._notebook.get_nth_page(page_num))

        widget.diagram_page.close()
        self._update_action_state()

    @event_handler(ModelReady)
    def _on_model_ready(self, event=None):
        """Open the toplevel element and load toplevel diagrams."""
        diagram_ids = self.properties.get("opened-diagrams", [])
        current_diagram_id = self.properties.get("current-diagram", None)

        diagrams = [self.element_factory.lookup(id) for id in diagram_ids]
        if not any(diagrams):
            diagrams = self.element_factory.select(
                lambda e: e.isKindOf(Diagram) and not (e.owner and e.owner.owner)
            )
        for diagram in diagrams:
            if diagram:
                self.event_manager.handle(DiagramOpened(diagram))

        if current_diagram_id:
            current_diagram = self.element_factory.lookup(current_diagram_id)
            if self.set_current_diagram(current_diagram):
                return
        if self._notebook and self._notebook.get_n_pages():
            if Gtk.get_major_version() == 3:
                self._notebook.set_current_page(0)
            else:
                self._notebook.set_selected_page(self._notebook.get_nth_page(0))

    @event_handler(ModelFlushed)
    def _on_flush_model(self, event):
        """Close all tabs."""
        if Gtk.get_major_version() == 3:
            while self._notebook.get_n_pages():
                self._notebook.remove_page(0)
        else:
            for page in self._notebook.get_pages():
                self._notebook.close_page(page)
        self._update_action_state()

    def _update_action_state(self):
        enabled = self._notebook and self._notebook.get_n_pages() > 0

        for action_name in ["win.zoom-in", "win.zoom-out", "win.zoom-100"]:
            self.event_manager.handle(ActionEnabled(action_name, enabled))

    @event_handler(AttributeUpdated)
    def _on_name_change(self, event):
        if event.property is Diagram.name:
            for page in range(self._notebook.get_n_pages()):
                widget = self._notebook.get_nth_page(page)
                if (
                    Gtk.get_major_version() == 3
                    and event.element is widget.diagram_page.diagram
                ):
                    self._notebook.set_tab_label(
                        widget, tab_label(event.new_value, widget, self.event_manager)
                    )
                    return
                elif event.element is widget.get_child().diagram_page.diagram:
                    widget.set_title(event.new_value or "")
                    return


def apply_tool_select_controller(widget, toolbox):
    if Gtk.get_major_version() == 3:
        ctrl = Gtk.EventControllerKey.new(widget)
        widget._toolbox_controller = ctrl
    else:
        ctrl = Gtk.EventControllerKey.new()
        widget.add_controller(ctrl)

    def on_shortcut(_ctrl, keyval, _keycode, state):
        toolbox.activate_shortcut(keyval, state)

    ctrl.connect("key-pressed", on_shortcut)


def tab_label(title, widget, event_manager):
    tab_box = Gtk.Box.new(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
    button = Gtk.Button()
    button.get_style_context().add_class("flat")
    button.set_focus_on_click(False)

    button.connect(
        "clicked",
        lambda _button: event_manager.handle(
            DiagramClosed(widget.diagram_page.get_diagram())
        ),
    )

    label = Gtk.Label.new(title)
    if Gtk.get_major_version() == 3:
        tab_box.pack_start(child=label, expand=True, fill=True, padding=0)
        close_image = Gtk.Image.new_from_icon_name(
            icon_name="window-close-symbolic", size=Gtk.IconSize.BUTTON
        )
        button.add(close_image)
        tab_box.pack_start(child=button, expand=False, fill=False, padding=0)
        tab_box.show_all()
    else:
        tab_box.append(label)
        close_image = Gtk.Image.new_from_icon_name("window-close-symbolic")
        button.set_child(close_image)
        tab_box.append(button)
        tab_box.show()

    return tab_box


def get_widgets_on_pages(notebook):
    """Gets the widget on each open page Notebook page.

    The page is the page number in the Notebook (0 indexed) and the widget
    is the child widget on each page.

    Returns:
        List of tuples (page, widget) of the currently open Notebook pages.
    """
    if not notebook:
        return

    if Gtk.get_major_version() == 3:
        for page_num in range(notebook.get_n_pages()):
            yield (page_num, notebook.get_nth_page(page_num))
    else:
        for page_num in range(notebook.get_n_pages()):
            yield (page_num, notebook.get_nth_page(page_num).get_child())



"""Final state diagram item."""

from gaphor import UML
from gaphor.diagram.presentation import ElementPresentation, Named
from gaphor.diagram.shapes import Box, IconBox, Text, ellipse, stroke
from gaphor.diagram.support import represents
from gaphor.UML.recipes import stereotypes_str


@represents(UML.FinalState)
class FinalStateItem(ElementPresentation, Named):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=30, height=30)
        for h in self.handles():
            h.movable = False

        self.shape = IconBox(
            Box(draw=draw_final_state),
            Text(
                text=lambda: stereotypes_str(self.subject),
            ),
            Text(text=lambda: self.subject and self.subject.name or ""),
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject.appliedStereotype.classifier.name")


def draw_final_state(box, context, bounding_box):
    cr = context.cairo
    r = 15

    d = r * 2
    ellipse(cr, *bounding_box)
    cr.set_line_width(0.01)
    cr.set_line_width(2)
    stroke(context)

    if stroke_color := context.style["color"]:
        cr.set_source_rgba(*stroke_color)

    d = 20
    ellipse(cr, 5, 5, d, d)
    cr.set_line_width(0.01)
    cr.fill()
